import argparse
import textwrap
import copy
import pandas as pd
import numpy as np
import yaml
from typing import List

from dingo.gw.SVD import SVDBasis
from dingo.gw.dataset.generate_dataset import train_svd_basis
from dingo.gw.dataset.waveform_dataset import WaveformDataset


def merge_datasets(dataset_list: List[WaveformDataset]) -> WaveformDataset:
    """
    Merge a collection of datasets into one.

    Parameters
    ----------
    dataset_list : list[WaveformDataset]
        A list of WaveformDatasets. Each item should be a dictionary containing
        parameters and polarizations.

    Returns
    -------
    WaveformDataset containing the merged data.
    """

    print(f"Merging {len(dataset_list)} datasets into one.")

    # This ensures that all the keys are copied into the new dataset. The "extensive"
    # parts of the dataset (parameters, waveforms) will be overwritten by the combined
    # datasets, whereas the "intensive" parts (e.g., SVD basis, settings) will take the
    # values in the *first* dataset in the list.
    merged_dict = copy.deepcopy(dataset_list[0].to_dictionary())

    merged_dict["parameters"] = pd.concat([d.parameters for d in dataset_list])
    merged_dict["polarizations"] = {}
    polarizations = list(dataset_list[0].polarizations.keys())
    for pol in polarizations:
        # We pop the data array off of each of the polarizations dicts to save memory.
        # Otherwise, this operation doubles the total amount of memory used. This is
        # destructive to the original datasets.
        merged_dict["polarizations"][pol] = np.vstack(
            [d.polarizations.pop(pol) for d in dataset_list]
        )

    # Update the settings based on the total number of samples.
    merged_dict["settings"]["num_samples"] = len(merged_dict["parameters"])

    merged = WaveformDataset(dictionary=merged_dict)

    return merged


def merge_datasets_cli():
    """
    Command-line function to combine a collection of datasets into one. Used for
    parallelized waveform generation.
    """

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent(
            """\
        Combine a collection of datasets into one.
        
        Datasets must be in sequentially-labeled HDF5 files with a fixed prefix. 
        The settings for the new dataset will be based on those of the first file. 
        Optionally, replace the settings with those specified in a YAML file.
        """
        ),
    )
    parser.add_argument(
        "--prefix", type=str, required=True, help="Prefix of sequential files names."
    )
    parser.add_argument(
        "--num_parts",
        type=int,
        required=True,
        help="Total number of datasets to merge.",
    )
    parser.add_argument(
        "--out_file", type=str, required=True, help="Name of file for new dataset."
    )
    parser.add_argument(
        "--settings_file", type=str, help="YAML file containing new dataset settings."
    )
    args = parser.parse_args()

    dataset_list = []
    for i in range(args.num_parts):
        file_name = args.prefix + str(i) + ".hdf5"
        dataset_list.append(WaveformDataset(file_name=file_name))
    merged_dataset = merge_datasets(dataset_list)

    # Optionally, update the settings file based on that provided at command line.
    if args.settings_file is not None:
        with open(args.settings_file, "r") as f:
            settings = yaml.safe_load(f)
        # Make sure num_samples is correct
        settings["num_samples"] = len(merged_dataset.parameters)
        merged_dataset.settings = settings

    merged_dataset.to_file(args.out_file)
    print(
        f"Complete. New dataset consists of {merged_dataset.settings['num_samples']} "
        f"samples."
    )


def build_svd_cli():
    """
    Command-line function to build an SVD based on an uncompressed dataset file.
    """

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent(
            """\
        Build an SVD basis based on a set of waveforms.
        """
        ),
    )
    parser.add_argument(
        "--dataset_file",
        type=str,
        required=True,
        help="HDF5 file containing training waveforms.",
    )
    parser.add_argument(
        "--size", type=int, required=True, help="Number of basis elements to keep."
    )
    parser.add_argument(
        "--out_file", type=str, required=True, help="Name of file for saving SVD."
    )
    parser.add_argument(
        "--num_train",
        type=int,
        help="Number of waveforms to use for training SVD. "
        "Remainder are used for validation.",
    )
    args = parser.parse_args()

    dataset = WaveformDataset(file_name=args.dataset_file)
    if args.num_train is None:
        n_train = len(WaveformDataset)
    else:
        n_train = args.num_train

    basis, n_train, n_test = train_svd_basis(dataset, args.size, n_train)
    # FIXME: This is not an ideal treatment. We should update the waveform generation
    #  to always provide the requested number of waveforms.
    print(
        f"SVD basis trained based on {n_train} waveforms and validated on {n_test} "
        f"waveforms. Note that if this differs from number requested, it will not be "
        f"reflected in the settings file. This is likely due to EOB failure to "
        f"generate certain waveforms."
    )
    basis.to_file(args.out_file)



import numpy as np
import torch


class PostCorrectGeocentTime(object):
    """
    Post correction for geocent time: add GNPE proxy (only necessary if exact
    equivariance is enforced)
    """

    def __init__(self, inverse=False):
        self.inverse = inverse

    def __call__(self, input_sample):
        sign = (1, -1)[self.inverse]
        sample = input_sample.copy()
        parameters = sample["parameters"].copy()
        extrinsic_parameters = sample["extrinsic_parameters"].copy()
        parameters["geocent_time"] -= extrinsic_parameters.pop("geocent_time") * sign
        sample["parameters"] = parameters
        sample["extrinsic_parameters"] = extrinsic_parameters
        return sample


class CopyToExtrinsicParameters(object):
    """
    Copy parameters specified in self.parameter_list from sample["parameters"] to
    sample["extrinsic_parameters"].
    """

    def __init__(self, *parameter_list):
        self.parameter_list = parameter_list

    def __call__(self, input_sample):
        sample = input_sample.copy()
        extrinsic_parameters = sample["extrinsic_parameters"].copy()
        for par in self.parameter_list:
            if par in sample['parameters']:
                extrinsic_parameters[par] = sample["parameters"][par]
        sample["extrinsic_parameters"] = extrinsic_parameters
        return sample


class ExpandStrain(object):
    """
    Expand the waveform of sample by adding a batch axis and copying the waveform
    num_samples times along this new axis. This is useful for generating num_samples
    samples at inference time.
    """

    def __init__(self, num_samples):
        self.num_samples = num_samples

    def __call__(self, input_sample):
        sample = input_sample.copy()
        waveform = input_sample["waveform"]
        sample["waveform"] = waveform.expand(self.num_samples, *waveform.shape)
        return sample


class ToTorch(object):
    """
    Convert all numpy arrays sample to torch tensors and push them to the specified
    device. All items of sample that are not numpy arrays (e.g., dicts of arrays)
    remain unchanged.
    """

    def __init__(self, device="cpu"):
        self.device = device

    def __call__(self, input_sample):
        sample = input_sample.copy()
        for k, v in sample.items():
            if type(v) == np.ndarray:
                sample[k] = torch.from_numpy(v).to(self.device, non_blocking=True)
        return sample


class ResetSample(object):
    """
    Resets sample:
        * waveform was potentially modified by gnpe transforms, so reset to waveform_
        * optionally remove all non-required extrinsic parameters
    """

    def __init__(self, extrinsic_parameters_keys=None):
        self.extrinsic_parameters_keys = extrinsic_parameters_keys

    def __call__(self, input_sample):
        sample = input_sample.copy()
        # reset the waveform
        sample["waveform"] = sample["waveform_"].clone()
        # optionally remove all non-required extrinsic parameters
        if self.extrinsic_parameters_keys is not None:
            sample["extrinsic_parameters"] = {
                k: sample["extrinsic_parameters"][k]
                for k in self.extrinsic_parameters_keys
            }
        return sample



def _strip_unwanted_submission_keys(job):
    job.getenv = None
    job.universe = None
    extra_lines = []
    for line in job.extra_lines:
        if not line.startswith("priority") and not line.startswith("accounting_group"):
            extra_lines.append(line)
    job.extra_lines = extra_lines



from typing import Dict

from functools import lru_cache
from abc import ABC, abstractmethod

import numpy as np
import torch

from dingo.gw.gwutils import *


class Domain(ABC):
    """Defines the physical domain on which the data of interest live.

    This includes a specification of the bins or points,
    and a few additional properties associated with the data.
    """

    @abstractmethod
    def __len__(self):
        """Number of bins or points in the domain"""
        pass

    @abstractmethod
    def __call__(self, *args, **kwargs) -> np.ndarray:
        """Array of bins in the domain"""
        pass

    @abstractmethod
    def update(self, new_settings: dict):
        pass

    @abstractmethod
    def time_translate_data(self, data, dt) -> np.ndarray:
        """Time translate strain data by dt seconds."""
        pass

    @property
    @abstractmethod
    def noise_std(self) -> float:
        """Standard deviation of the whitened noise distribution"""
        # FIXME: For this to make sense, it assumes knowledge about how the domain is used in conjunction
        #  with (waveform) data, whitening and adding noise. Is this the best place to define this?
        pass

    @property
    @abstractmethod
    def sampling_rate(self) -> float:
        """The sampling rate of the data [Hz]."""
        pass

    @property
    @abstractmethod
    def f_max(self) -> float:
        """The maximum frequency [Hz] is set to half the sampling rate."""

    @property
    @abstractmethod
    def duration(self) -> float:
        """Waveform duration in seconds."""
        pass

    @property
    @abstractmethod
    def min_idx(self) -> int:
        pass

    @property
    @abstractmethod
    def max_idx(self) -> int:
        pass

    @property
    @abstractmethod
    def domain_dict(self):
        """Enables to rebuild the domain via calling build_domain(domain_dict)."""
        pass

    def __eq__(self, other):
        if self.domain_dict == other.domain_dict:
            return True
        else:
            return False


class FrequencyDomain(Domain):
    """Defines the physical domain on which the data of interest live.

    The frequency bins are assumed to be uniform between [0, f_max]
    with spacing delta_f.
    Given a finite length of time domain data, the Fourier domain data
    starts at a frequency f_min and is zero below this frequency.
    window_kwargs specify windowing used for FFT to obtain FD data from TD
    data in practice.
    """

    def __init__(
        self, f_min: float, f_max: float, delta_f: float, window_factor: float = None
    ):
        self._f_min = f_min
        self._f_max = f_max
        self._delta_f = delta_f
        self._window_factor = window_factor

        self._sample_frequencies = None
        self._sample_frequencies_torch = None
        self._sample_frequencies_torch_cuda = None
        self._frequency_mask = None

    def update(self, new_settings: dict):
        """
        Update the domain with new settings. This is only allowed if the new settings
        are "compatible" with the old ones. E.g., f_min should be larger than the
        existing f_min.

        Parameters
        ----------
        new_settings : dict
            Settings dictionary. Must contain a subset of the keys contained in
            domain_dict.
        """
        new_settings = new_settings.copy()
        if "type" in new_settings and new_settings.pop("type") not in [
            "FrequencyDomain",
            "FD",
        ]:
            raise ValueError("Cannot update domain to type other than FrequencyDomain.")
        for k, v in new_settings.items():
            if k not in ["f_min", "f_max", "delta_f", "window_factor"]:
                raise KeyError(f"Invalid key for domain update: {k}.")
            if k == "window_factor" and v != self._window_factor:
                raise ValueError("Cannot update window_factor.")
            if k == "delta_f" and v != self._delta_f:
                raise ValueError("Cannot update delta_f.")
        self.set_new_range(
            f_min=new_settings.get("f_min", None), f_max=new_settings.get("f_max", None)
        )

    def set_new_range(self, f_min: float = None, f_max: float = None):
        """
        Set a new range [f_min, f_max] for the domain. This is only allowed if the new
        range is contained within the old one.
        """
        if f_min is not None and f_max is not None and f_min >= f_max:
            raise ValueError("f_min must not be larger than f_max.")
        if f_min is not None:
            if self.f_min <= f_min <= self.f_max:
                self.f_min = f_min
            else:
                raise ValueError(
                    f"f_min = {f_min} is not in expected range "
                    f"[{self.f_min,self.f_max}]."
                )
        if f_max is not None:
            if self.f_min <= f_max <= self.f_max:
                self.f_max = f_max
            else:
                raise ValueError(
                    f"f_max = {f_max} is not in expected range "
                    f"[{self.f_min, self.f_max}]."
                )

    def update_data(self, data: np.ndarray, axis: int = -1, low_value: float = 0.0):
        """
        Adjusts data to be compatible with the domain:

        * Below f_min, it sets the data to low_value (typically 0.0 for a waveform,
          but for a PSD this might be a large value).
        * Above f_max, it truncates the data array.

        Parameters
        ----------
        data : np.ndarray
            Data array
        axis : int
            Which data axis to apply the adjustment along.
        low_value : float
            Below f_min, set the data to this value.

        Returns
        -------
        np.ndarray
            The new data array.
        """
        sl = [slice(None)] * data.ndim

        # First truncate beyond f_max.
        sl[axis] = slice(0, self.max_idx + 1)
        data = data[tuple(sl)]

        # Set data value below f_min to low_value.
        sl[axis] = slice(0, self.min_idx)
        data[tuple(sl)] = low_value

        return data

    def time_translate_data(self, data, dt):
        """
        Time translate frequency-domain data by dt. Time translation corresponds (in
        frequency domain) to multiplication by

        .. math::
            \exp(-2 \pi i \, f \, dt).

        This method allows for multiple batch dimensions. For torch.Tensor data,
        allow for either a complex or a (real, imag) representation.

        Parameters
        ----------
        data : array-like (numpy, torch)
            Shape (B, C, N), where

                - B corresponds to any dimension >= 0,
                - C is either absent (for complex data) or has dimension >= 2 (for data
                  represented as real and imaginary parts), and
                - N is either len(self) or len(self)-self.min_idx (for truncated data).

        dt : torch tensor, or scalar (if data is numpy)
            Shape (B)

        Returns
        -------
        Array-like of the same form as data.
        """
        f = self.get_sample_frequencies_astype(data)
        if isinstance(data, np.ndarray):
            # Assume numpy arrays un-batched, since they are only used at train time.
            phase_shift = 2 * np.pi * dt * f
        elif isinstance(data, torch.Tensor):
            # Allow for possible multiple "batch" dimensions (e.g., batch + detector,
            # which might have independent time shifts).
            phase_shift = 2 * np.pi * torch.einsum("...,i", dt, f)
        else:
            raise NotImplementedError(
                f"Time translation not implemented for data of " "type {data}."
            )
        return self.add_phase(data, phase_shift)

    def get_sample_frequencies_astype(self, data):
        """
        Returns a 1D frequency array compatible with the last index of data array.

        Decides whether array is numpy or torch tensor (and cuda vs cpu), and whether it
        contains the leading zeros below f_min.

        Parameters
        ----------
        data : Union[np.array, torch.Tensor]
            Sample data

        Returns
        -------
        frequency array compatible with last index
        """
        # Type
        if isinstance(data, np.ndarray):
            f = self.sample_frequencies
        elif isinstance(data, torch.Tensor):
            if data.is_cuda:
                f = self.sample_frequencies_torch_cuda
            else:
                f = self.sample_frequencies_torch
        else:
            raise TypeError("Invalid data type. Should be np.array or torch.Tensor.")

        # Whether to include zeros below f_min
        if data.shape[-1] == len(self) - self.min_idx:
            f = f[self.min_idx :]
        elif data.shape[-1] != len(self):
            raise TypeError(
                f"Data with {data.shape[-1]} frequency bins is "
                f"incompatible with domain."
            )

        return f

    @staticmethod
    def add_phase(data, phase):
        """
        Add a (frequency-dependent) phase to a frequency series. Allows for batching,
        as well as additional channels (such as detectors). Accounts for the fact that
        the data could be a complex frequency series or real and imaginary parts.

        Convention: the phase phi(f) is defined via exp(- 1j * phi(f)).

        Parameters
        ----------
        data : Union[np.array, torch.Tensor]
        phase : Union[np.array, torch.Tensor]

        Returns
        -------
        New array or tensor of the same shape as data.
        """
        if isinstance(data, np.ndarray) and np.iscomplexobj(data):
            # This case is assumed to only occur during inference, with un-batched data.
            return data * np.exp(-1j * phase)

        elif isinstance(data, torch.Tensor):
            if torch.is_complex(data):
                # Expand the trailing batch dimensions to allow for broadcasting.
                while phase.dim() < data.dim():
                    phase = phase[..., None, :]
                return data * torch.exp(-1j * phase)
            else:
                # The first two components of the second last index should be the real
                # and imaginary parts of the data. Remaining components correspond to,
                # e.g., the ASD. The "-1" below accounts for this extra dimension when
                # broadcasting.
                while phase.dim() < data.dim() - 1:
                    phase = phase[..., None, :]

                cos_phase = torch.cos(phase)
                sin_phase = torch.sin(phase)
                result = torch.empty_like(data)
                result[..., 0, :] = (
                    data[..., 0, :] * cos_phase + data[..., 1, :] * sin_phase
                )
                result[..., 1, :] = (
                    data[..., 1, :] * cos_phase - data[..., 0, :] * sin_phase
                )
                if data.shape[-2] > 2:
                    result[..., 2:, :] = data[..., 2:, :]
                return result

        else:
            raise TypeError(f"Invalid data type {type(data)}.")

    def __len__(self):
        """Number of frequency bins in the domain [0, f_max]"""
        return int(self.f_max / self.delta_f) + 1

    def __call__(self) -> np.ndarray:
        """Array of uniform frequency bins in the domain [0, f_max]"""
        return self.sample_frequencies

    def __getitem__(self, idx):
        """Slice of uniform frequency grid."""
        sample_frequencies = self.__call__()
        return sample_frequencies[idx]

    @property
    def sample_frequencies(self):
        if self._sample_frequencies is None:
            num_bins = len(self)
            self._sample_frequencies = np.linspace(
                0.0, self.f_max, num=num_bins, endpoint=True, dtype=np.float32
            )
        return self._sample_frequencies

    @property
    def sample_frequencies_torch(self):
        if self._sample_frequencies_torch is None:
            num_bins = len(self)
            self._sample_frequencies_torch = torch.linspace(
                0.0, self.f_max, steps=num_bins, dtype=torch.float32
            )
        return self._sample_frequencies_torch

    @property
    def sample_frequencies_torch_cuda(self):
        if self._sample_frequencies_torch_cuda is None:
            self._sample_frequencies_torch_cuda = self.sample_frequencies_torch.to(
                "cuda"
            )
        return self._sample_frequencies_torch_cuda

    @property
    def frequency_mask(self) -> np.ndarray:
        """Mask which selects frequency bins greater than or equal to the
        starting frequency"""
        if self._frequency_mask is None:
            self._frequency_mask = self.sample_frequencies >= self.f_min
        return self._frequency_mask

    def _reset_caches(self):
        self._sample_frequencies = None
        self._sample_frequencies_torch = None
        self._sample_frequencies_torch_cuda = None
        self._frequency_mask = None

    @property
    def frequency_mask_length(self) -> int:
        """Number of samples in the subdomain domain[frequency_mask]."""
        mask = self.frequency_mask
        return len(np.flatnonzero(np.asarray(mask)))

    @property
    def min_idx(self):
        return round(self._f_min / self._delta_f)

    @property
    def max_idx(self):
        return round(self._f_max / self._delta_f)

    @property
    def window_factor(self):
        return self._window_factor

    @window_factor.setter
    def window_factor(self, value):
        """Set self._window_factor and clear cache of self.noise_std."""
        self._window_factor = float(value)

    @property
    def noise_std(self) -> float:
        """Standard deviation of the whitened noise distribution.

        To have noise that comes from a multivariate *unit* normal
        distribution, you must divide by this factor. In practice, this means
        dividing the whitened waveforms by this.

        TODO: This description makes some assumptions that need to be clarified.
        Windowing of TD data; tapering window has a slope -> reduces power only for noise,
        but not for the signal which is in the main part unaffected by the taper
        """
        if self._window_factor is None:
            raise ValueError("Window factor needs to be set for noise_std.")
        return np.sqrt(self._window_factor) / np.sqrt(4.0 * self._delta_f)

    @property
    def f_max(self) -> float:
        """The maximum frequency [Hz] is typically set to half the sampling
        rate."""
        return self._f_max

    @f_max.setter
    def f_max(self, value):
        self._f_max = float(value)
        self._reset_caches()

    @property
    def f_min(self) -> float:
        """The minimum frequency [Hz]."""
        return self._f_min

    @f_min.setter
    def f_min(self, value):
        self._f_min = float(value)
        self._reset_caches()

    @property
    def delta_f(self) -> float:
        """The frequency spacing of the uniform grid [Hz]."""
        return self._delta_f

    @delta_f.setter
    def delta_f(self, value):
        self._delta_f = float(value)
        self._reset_caches()

    @property
    def duration(self) -> float:
        """Waveform duration in seconds."""
        return 1.0 / self.delta_f

    @property
    def sampling_rate(self) -> float:
        return 2.0 * self.f_max

    @property
    def domain_dict(self):
        """Enables to rebuild the domain via calling build_domain(domain_dict)."""
        return {
            "type": "FrequencyDomain",
            "f_min": self.f_min,
            "f_max": self.f_max,
            "delta_f": self.delta_f,
            "window_factor": self.window_factor,
        }


class TimeDomain(Domain):
    """Defines the physical time domain on which the data of interest live.

    The time bins are assumed to be uniform between [0, duration]
    with spacing 1 / sampling_rate.
    window_factor is used to compute noise_std().
    """

    def __init__(self, time_duration: float, sampling_rate: float):
        self._time_duration = time_duration
        self._sampling_rate = sampling_rate

    @lru_cache()
    def __len__(self):
        """Number of time bins given duration and sampling rate"""
        return int(self._time_duration * self._sampling_rate)

    @lru_cache()
    def __call__(self) -> np.ndarray:
        """Array of uniform times at which data is sampled"""
        num_bins = self.__len__()
        return np.linspace(
            0.0, self._time_duration, num=num_bins, endpoint=False, dtype=np.float32
        )

    @property
    def delta_t(self) -> float:
        """The size of the time bins"""
        return 1.0 / self._sampling_rate

    @delta_t.setter
    def delta_t(self, delta_t: float):
        self._sampling_rate = 1.0 / delta_t

    @property
    def noise_std(self) -> float:
        """Standard deviation of the whitened noise distribution.

        To have noise that comes from a multivariate *unit* normal
        distribution, you must divide by this factor. In practice, this means
        dividing the whitened waveforms by this.

        In the continuum limit in time domain, the standard deviation of white
        noise would at each point go to infinity, hence the delta_t factor.
        """
        return 1.0 / np.sqrt(2.0 * self.delta_t)

    def time_translate_data(self, data, dt) -> np.ndarray:
        raise NotImplementedError

    @property
    def f_max(self) -> float:
        """The maximum frequency [Hz] is typically set to half the sampling
        rate."""
        return self._sampling_rate / 2.0

    @property
    def duration(self) -> float:
        """Waveform duration in seconds."""
        return self._time_duration

    @property
    def sampling_rate(self) -> float:
        return self._sampling_rate

    @property
    def min_idx(self) -> int:
        return 0

    @property
    def max_idx(self) -> int:
        return round(self._time_duration * self._sampling_rate)

    @property
    def domain_dict(self):
        """Enables to rebuild the domain via calling build_domain(domain_dict)."""
        return {
            "type": "TimeDomain",
            "time_duration": self._time_duration,
            "sampling_rate": self._sampling_rate,
        }


class PCADomain(Domain):
    """TODO"""

    # Not super important right now
    # FIXME: Should this be defined for FD or TD bases or both?
    # Nrb instead of Nf

    @property
    def noise_std(self) -> float:
        """Standard deviation of the whitened noise distribution.

        To have noise that comes from a multivariate *unit* normal
        distribution, you must divide by this factor. In practice, this means
        dividing the whitened waveforms by this.

        In the continuum limit in time domain, the standard deviation of white
        noise would at each point go to infinity, hence the delta_t factor.
        """
        # FIXME
        return np.sqrt(self.window_factor) / np.sqrt(4.0 * self.delta_f)


def build_domain(settings: Dict) -> Domain:
    """
    Instantiate a domain class from settings.

    Parameters
    ----------
    settings : dict
        Dicionary with 'type' key denoting the type of domain, and keys corresponding
        to the kwargs needed to construct the Domain.

    Returns
    -------
    A Domain instance of the correct type.
    """
    if "type" not in settings:
        raise ValueError(
            f'Domain settings must include a "type" key. Settings included '
            f"the keys {settings.keys()}."
        )

    # The settings other than 'type' correspond to the kwargs of the Domain constructor.
    kwargs = {k: v for k, v in settings.items() if k != "type"}
    if settings["type"] in ["FrequencyDomain", "FD"]:
        return FrequencyDomain(**kwargs)
    elif settings["type"] == ["TimeDomain", "TD"]:
        return TimeDomain(**kwargs)
    else:
        raise NotImplementedError(f'Domain {settings["name"]} not implemented.')


def build_domain_from_model_metadata(model_metadata) -> Domain:
    """
    Instantiate a domain class from settings of model.

    Parameters
    ----------
    model_metadata: dict
        model metadata containing information to build the domain
        typically obtained from the model.metadata attribute

    Returns
    -------
    A Domain instance of the correct type.
    """
    domain = build_domain(model_metadata["dataset_settings"]["domain"])
    if "domain_update" in model_metadata["train_settings"]["data"]:
        domain.update(model_metadata["train_settings"]["data"]["domain_update"])
    domain.window_factor = get_window_factor(
        model_metadata["train_settings"]["data"]["window"]
    )
    return domain


if __name__ == "__main__":
    kwargs = {"f_min": 20, "f_max": 2048, "delta_f": 0.125}
    domain = FrequencyDomain(**kwargs)

    d1 = domain()
    d2 = domain()
    print("Clearing cache.", end=" ")
    domain.clear_cache_for_all_instances()
    print("Done.")
    d3 = domain()

    print("Changing domain range.", end=" ")
    domain.set_new_range(20, 100)
    print("Done.")

    d4 = domain()
    d5 = domain()

    print(len(d1), len(d4))



import os
import sys

from bilby_pipe.input import Input
from bilby_pipe.main import parse_args
from bilby_pipe.utils import logger, convert_string_to_dict
from bilby_pipe.data_generation import DataGenerationInput as BilbyDataGenerationInput

from dingo.gw.data.event_dataset import EventDataset
from dingo.gw.domains import FrequencyDomain
from dingo.pipe.parser import create_parser

logger.name = "dingo_pipe"


class DataGenerationInput(BilbyDataGenerationInput):
    def __init__(self, args, unknown_args, create_data=True):
        Input.__init__(self, args, unknown_args)

        # Generic initialisation
        self.meta_data = dict(
            command_line_args=args.__dict__,
            unknown_command_line_args=unknown_args,
            injection_parameters=None,
            # bilby_version=bilby.__version__,
            # bilby_pipe_version=get_version_information(),
        )
        self.injection_parameters = None

        # Admin arguments
        self.ini = args.ini

        # Run index arguments
        self.idx = args.idx
        # self.generation_seed = args.generation_seed
        self.trigger_time = args.trigger_time

        # Naming arguments
        self.outdir = args.outdir
        self.label = args.label

        # Prior arguments
        # self.reference_frame = args.reference_frame
        # self.time_reference = args.time_reference
        # self.phase_marginalization = args.phase_marginalization
        # self.prior_file = args.prior_file
        # self.prior_dict = args.prior_dict
        # self.deltaT = args.deltaT
        # self.default_prior = args.default_prior

        # Whether to generate data for importance sampling. This must be done when
        # desired data settings differ from those used for network training. If this is
        # the case, save the new data to a different file name.

        self.importance_sampling = args.importance_sampling_generation
        self.importance_sampling_updates = args.importance_sampling_updates
        if self.importance_sampling:
            vars(args).update(self.importance_sampling_updates)

        # Data arguments
        self.ignore_gwpy_data_quality_check = args.ignore_gwpy_data_quality_check
        self.detectors = args.detectors
        self.channel_dict = args.channel_dict
        self.data_dict = args.data_dict
        self.data_format = args.data_format
        self.allow_tape = args.allow_tape
        self.tukey_roll_off = args.tukey_roll_off
        self.zero_noise = False  # dingo mod
        self.resampling_method = args.resampling_method

        # if args.timeslide_dict is not None:
        #     self.timeslide_dict = convert_string_to_dict(args.timeslide_dict)
        #     logger.info(f"Read-in timeslide dict directly: {self.timeslide_dict}")
        # elif args.timeslide_file is not None:
        #     self.gps_file = args.gps_file
        #     self.timeslide_file = args.timeslide_file
        #     self.timeslide_dict = self.get_timeslide_dict(self.idx)

        # Data duration arguments
        self.duration = args.duration
        self.post_trigger_duration = args.post_trigger_duration

        # Frequencies
        self.sampling_frequency = args.sampling_frequency
        self.minimum_frequency = args.minimum_frequency
        self.maximum_frequency = args.maximum_frequency
        # self.reference_frequency = args.reference_frequency

        # Waveform, source model and likelihood
        # self.waveform_generator_class = args.waveform_generator
        # self.waveform_approximant = args.waveform_approximant
        # self.catch_waveform_errors = args.catch_waveform_errors
        # self.pn_spin_order = args.pn_spin_order
        # self.pn_tidal_order = args.pn_tidal_order
        # self.pn_phase_order = args.pn_phase_order
        # self.pn_amplitude_order = args.pn_amplitude_order
        # self.mode_array = args.mode_array
        # self.waveform_arguments_dict = args.waveform_arguments_dict
        # self.numerical_relativity_file = args.numerical_relativity_file
        # self.injection_waveform_approximant = args.injection_waveform_approximant
        # self.frequency_domain_source_model = args.frequency_domain_source_model
        # self.conversion_function = args.conversion_function
        # self.generation_function = args.generation_function
        # self.likelihood_type = args.likelihood_type
        # self.extra_likelihood_kwargs = args.extra_likelihood_kwargs
        # self.enforce_signal_duration = args.enforce_signal_duration

        # PSD
        self.psd_maximum_duration = args.psd_maximum_duration
        self.psd_dict = args.psd_dict
        if self.psd_dict is None:
            self.psd_length = args.psd_length
            self.psd_fractional_overlap = args.psd_fractional_overlap
            self.psd_start_time = args.psd_start_time
            self.psd_method = args.psd_method

        # # ROQ
        # self.roq_folder = args.roq_folder
        # self.roq_linear_matrix = args.roq_linear_matrix
        # self.roq_quadratic_matrix = args.roq_quadratic_matrix
        # self.roq_weights = args.roq_weights
        # self.roq_weight_format = args.roq_weight_format
        # self.roq_scale_factor = args.roq_scale_factor
        #
        # # Calibration
        # self.calibration_model = args.calibration_model
        # self.spline_calibration_envelope_dict = args.spline_calibration_envelope_dict
        # self.spline_calibration_amplitude_uncertainty_dict = (
        #     args.spline_calibration_amplitude_uncertainty_dict
        # )
        # self.spline_calibration_phase_uncertainty_dict = (
        #     args.spline_calibration_phase_uncertainty_dict
        # )
        # self.spline_calibration_nodes = args.spline_calibration_nodes
        # self.calibration_prior_boundary = args.calibration_prior_boundary
        #
        # # Marginalization
        # self.distance_marginalization = args.distance_marginalization
        # self.distance_marginalization_lookup_table = (
        #     args.distance_marginalization_lookup_table
        # )
        # self.phase_marginalization = args.phase_marginalization
        # self.time_marginalization = args.time_marginalization
        # self.jitter_time = args.jitter_time
        #
        # # Plotting
        self.plot_data = args.plot_data
        self.plot_spectrogram = args.plot_spectrogram
        # self.plot_injection = args.plot_injection

        if create_data:
            # Added for dingo so that create_data runs. TODO: enable injections
            args.injection = False
            args.injection_numbers = None
            args.injection_file = None
            args.injection_dict = None
            args.gaussian_noise = False
            args.injection_waveform_arguments = None

            self.create_data(args)

    def save_hdf5(self):
        """Save frequency-domain strain and ASDs as DingoDataset HDF5 format."""

        # PSD and strain data.
        data = {"waveform": {}, "asds": {}}  # TODO: Rename these keys.
        for ifo in self.interferometers:

            strain = ifo.strain_data.frequency_domain_strain
            frequency_array = ifo.strain_data.frequency_array
            asd = ifo.power_spectral_density.get_amplitude_spectral_density_array(
                frequency_array
            )

            # These arrays extend up to self.sampling_frequency. Truncate them to
            # self.maximum_frequency, and also set the asd to 1.0 below
            # self.minimum_frequency.
            domain = FrequencyDomain(
                f_min=self.minimum_frequency,
                f_max=self.maximum_frequency,
                delta_f=1 / self.duration,
            )
            strain = domain.update_data(strain)
            asd = domain.update_data(asd, low_value=1.0)

            # Dingo expects data to have trigger time 0, so we apply a cyclic time shift
            # by the post-trigger duration.
            strain = domain.time_translate_data(strain, self.post_trigger_duration)

            # Note that the ASD estimated by the Bilby Interferometer differs ever so
            # slightly from the ASDs we computed before using pycbc. In addition,
            # there is no handling of NaNs in the strain data from which the ASD is
            # estimated.

            data["waveform"][ifo.name] = strain
            data["asds"][ifo.name] = asd

        # Data conditioning settings.
        settings = {
            "time_event": self.trigger_time,
            "time_buffer": self.post_trigger_duration,
            "detectors": self.detectors,
            "f_s": self.sampling_frequency,
            "T": self.duration,
            "f_min": self.minimum_frequency,
            "f_max": self.maximum_frequency,
            "window_type": "tukey",
            "roll_off": self.tukey_roll_off,
        }

        for k in [
            "psd_duration",
            "psd_dict",
            "psd_fractional_overlap",
            "psd_start_time",
            "psd_method",
            "channel_dict",
            "data_dict",
        ]:
            try:
                v = getattr(self, k)
            except AttributeError:
                continue
            if v is not None:
                settings[k] = v

        dataset = EventDataset(
            dictionary={
                "data": data,
                # "event_metadata": event_metadata,
                "settings": settings,
            }
        )
        dataset.to_file(self.event_data_file)

    @property
    def event_data_file(self):
        return os.path.join(
            self.data_directory, "_".join([self.label, "event_data.hdf5"])
        )

    @property
    def importance_sampling_updates(self):
        return self._importance_sampling_updates

    @importance_sampling_updates.setter
    def importance_sampling_updates(self, setting):
        if setting is not None:
            self._importance_sampling_updates = convert_string_to_dict(
                setting, "importance-sampling-updates"
            )
        else:
            self._importance_sampling_updates = None


def create_generation_parser():
    """Data generation parser creation"""
    return create_parser(top_level=False)


def main():
    """Data generation main logic"""
    args, unknown_args = parse_args(sys.argv[1:], create_generation_parser())
    # log_version_information()
    data = DataGenerationInput(args, unknown_args)
    data.save_hdf5()
    logger.info("Completed data generation")



import pytest
import types
import os
from os.path import join
import numpy as np
import torch
from dingo.core.models.posterior_model import PosteriorModel
from dingo.core.utils import torchutils


@pytest.fixture()
def data_setup_pm_1():
    d = types.SimpleNamespace()

    tmp_dir = "./tmp_files"
    os.makedirs(tmp_dir, exist_ok=True)
    d.model_filename = join(tmp_dir, "model.pt")

    d.nsf_kwargs = {
        "input_dim": 4,
        "context_dim": 10,
        "num_flow_steps": 5,
        "base_transform_kwargs": {
            "hidden_dim": 64,
            "num_transform_blocks": 2,
            "activation": "elu",
            "dropout_probability": 0.0,
            "batch_norm": True,
            "num_bins": 8,
            "base_transform_type": "rq-coupling",
        },
    }
    d.embedding_net_kwargs = {
        "input_dims": (2, 3, 20),
        # 'n_rb': 10,
        "svd": {"size": 10},
        "V_rb_list": None,
        "output_dim": 8,
        "hidden_dims": [32, 16, 8],
        "activation": "elu",
        "dropout": 0.0,
        "batch_norm": True,
        "added_context": True,
    }

    d.model_kwargs = {
        "type": "nsf+embedding",
        "nsf_kwargs": d.nsf_kwargs,
        "embedding_net_kwargs": d.embedding_net_kwargs,
    }

    d.metadata = {
        "train_settings": {"model": d.model_kwargs}
    }

    return d


@pytest.fixture()
def data_setup_optimizer_scheduler():
    d = types.SimpleNamespace()

    d.adam_kwargs = {
        "type": "adam",
        "lr": 0.0001,
    }
    d.sgd_kwargs = {
        "type": "sgd",
        "lr": 0.0001,
        "momentum": 0.9,
    }

    d.step_kwargs = {"type": "step", "step_size": 3, "gamma": 0.5}
    d.cosine_kwargs = {
        "type": "cosine",
        "T_max": 10,
    }
    d.rop_kwargs = {
        "type": "reduce_on_plateau",
        "factor": 0.5,
        "patience": 3,
    }

    d.epochs = range(1, 10 + 1)
    d.losses = [-1, -2, -3, -2, -2, -2, -2, -7, -8, -9]
    d.rop_factors = [1, 1, 1, 1, 1, 1, 1, 0.5, 0.5, 0.5]
    d.cosine_factors = (1 + np.cos(np.linspace(0, np.pi, 11))[:-1]) / 2
    d.step_factors = [
        0.5 ** 0,
        0.5 ** 0,
        0.5 ** 0,
        0.5 ** 1,
        0.5 ** 1,
        0.5 ** 1,
        0.5 ** 2,
        0.5 ** 2,
        0.5 ** 2,
        0.5 ** 3,
    ]

    return d


def test_pm_saving_and_loading_basic(data_setup_pm_1):
    """
    Test the most basic functionality of initializing, saving and loading a
    model. Two models built with identical hyperparameters should have the
    same architecture (names and shapes of parameters), but they should be
    initialized differently. A model loaded from a saved file should have the
    exact same parameter data as the original model.
    """
    d = data_setup_pm_1

    # initialize model and save it
    pm_0 = PosteriorModel(metadata=d.metadata, device='cpu')
    pm_0.save_model(d.model_filename)

    # load saved model
    pm_1 = PosteriorModel(model_filename=d.model_filename, device='cpu')

    # build a model with identical kwargs
    pm_2 = PosteriorModel(metadata=d.metadata, device='cpu')

    # check that module names are identical in saved and loaded model
    module_names_0 = [name for name, param in pm_0.model.named_parameters()]
    module_names_1 = [name for name, param in pm_1.model.named_parameters()]
    module_names_2 = [name for name, param in pm_2.model.named_parameters()]
    assert (
        module_names_0 == module_names_1 == module_names_2
    ), "Models have different module names."

    # check that number of parameters are identical in saved and loaded model
    num_params_0 = torchutils.get_number_of_model_parameters(pm_0.model)
    num_params_1 = torchutils.get_number_of_model_parameters(pm_1.model)
    num_params_2 = torchutils.get_number_of_model_parameters(pm_2.model)
    assert (
        num_params_0 == num_params_1 == num_params_2
    ), "Models have different number of parameters."

    # check that the weights of pm_0 and pm_1 are identical, but different
    # compared to pm_2
    for name, param_0, param_1, param_2 in zip(
        module_names_0,
        pm_0.model.parameters(),
        pm_1.model.parameters(),
        pm_2.model.parameters(),
    ):
        assert (
            param_0.data.shape == param_1.data.shape == param_2.data.shape
        ), "Models have different parameter shapes."
        assert torch.all(
            param_0.data == param_1.data
        ), "Saved and loaded model have different parameter data."
        if not torch.std(param_0) == 0:
            assert not torch.all(
                param_0.data == param_2.data
            ), "Model initialization does not seem to be random."


def test_pm_scheduler(data_setup_pm_1, data_setup_optimizer_scheduler):
    """
    Test that the scheduler of PosteriorModel works as intended, also when
    saving and loading the model.
    """
    d = data_setup_pm_1
    e = data_setup_optimizer_scheduler

    # Test step scheduler
    optimizer_kwargs = e.adam_kwargs
    scheduler_kwargs = e.step_kwargs
    pm = PosteriorModel(metadata=d.metadata, device='cpu')
    pm.optimizer_kwargs = optimizer_kwargs
    pm.scheduler_kwargs = scheduler_kwargs
    pm.initialize_optimizer_and_scheduler()
    pm.optimizer.step()  # this is to suppress a pytorch warning
    factors = []
    for epoch, loss in zip(e.epochs, e.losses):
        if epoch == 5:
            pm.save_model(d.model_filename, save_training_info=True)
            pm = PosteriorModel(model_filename=d.model_filename, device='cpu')
        lr = pm.optimizer.state_dict()["param_groups"][0]["lr"]
        factors.append(lr / pm.optimizer.defaults["lr"])
        torchutils.perform_scheduler_step(pm.scheduler, loss)
    assert factors == e.step_factors, "Scheduler does not load correctly."

    # Test reduce_on_plateau scheduler
    optimizer_kwargs = e.sgd_kwargs
    scheduler_kwargs = e.rop_kwargs
    pm = PosteriorModel(metadata=d.metadata, device='cpu')
    pm.optimizer_kwargs = optimizer_kwargs
    pm.scheduler_kwargs = scheduler_kwargs
    pm.initialize_optimizer_and_scheduler()
    pm.optimizer.step()  # this is to suppress a pytorch warning
    factors = []
    for epoch, loss in zip(e.epochs, e.losses):
        if epoch == 5:
            pm.save_model(d.model_filename, save_training_info=True)
            pm = PosteriorModel(model_filename=d.model_filename, device='cpu')
        lr = pm.optimizer.state_dict()["param_groups"][0]["lr"]
        factors.append(lr / pm.optimizer.defaults["lr"])
        torchutils.perform_scheduler_step(pm.scheduler, loss)
    assert factors == e.rop_factors, "Scheduler does not load correctly."

    # Test cosine scheduler
    optimizer_kwargs = e.sgd_kwargs
    scheduler_kwargs = e.cosine_kwargs
    pm = PosteriorModel(metadata=d.metadata, device='cpu')
    pm.optimizer_kwargs = optimizer_kwargs
    pm.scheduler_kwargs = scheduler_kwargs
    pm.initialize_optimizer_and_scheduler()
    pm.optimizer.step()  # this is to suppress a pytorch warning
    factors = []
    for epoch, loss in zip(e.epochs, e.losses):
        if epoch == 5:
            pm.save_model(d.model_filename, save_training_info=True)
            pm = PosteriorModel(model_filename=d.model_filename, device='cpu')
        lr = pm.optimizer.state_dict()["param_groups"][0]["lr"]
        factors.append(lr / pm.optimizer.defaults["lr"])
        torchutils.perform_scheduler_step(pm.scheduler, loss)
    assert np.allclose(factors, e.cosine_factors), "Scheduler does not load correctly."



import argparse
import ast
import textwrap
from pathlib import Path
import h5py
import torch


def main():

    parser = argparse.ArgumentParser(
        description=textwrap.dedent(
            """Removes any domain window_factor from dataset or model settings. This 
        modifies the file in-place."""
        )
    )
    parser.add_argument(
        "file_name", type=str, help="Dataset or model file to be modified."
    )
    args = parser.parse_args()

    path = Path(args.file_name)

    if path.suffix == ".hdf5":
        with h5py.File(path, "r+") as f:
            settings = ast.literal_eval(f.attrs["settings"])
            try:
                if "domain" in settings:
                    del settings["domain"]["window_factor"]
                elif "domain_dict" in settings:
                    del settings["domain_dict"]["window_factor"]
                f.attrs["settings"] = str(settings)
            except KeyError:
                print("Dataset is already in correct format.")

    elif path.suffix == ".pt":
        if torch.cuda.is_available():
            d = torch.load(path)
        else:
            d = torch.load(path, map_location=torch.device("cpu"))

        try:
            del d["metadata"]["dataset_settings"]["domain"]["window_factor"]
            torch.save(d, path)
        except KeyError:
            print("Dataset is already in correct format.")


if __name__ == "__main__":
    main()



import os
import pickle
from os.path import join
from typing import Dict

import numpy as np
import yaml
from pycondor import Job, Dagman


def create_args_string(args_dict: Dict):
    """Generate argument string from dictionary of argument names and arguments."""
    return "".join([f"--{k} {v} " for k, v in args_dict.items()])


def split_time_segments(time_segments, condor_dir, num_jobs):
    """
    Split up all time segments used for estimating PSDs into num_jobs-many
    segments and save them into a condor directory

    Parameters
    ----------
    time_segments : dict
        contains all time segments used for estimating PSDs
    condor_dir : str
        path to a directory where condr-related files are stored
    num_jobs : int
        number of jobs that should be used per detector to parallelize the PSD estimation
    Returns
    -------
    List of paths where the files including the subsets of all time segments are stored
    """
    time_segments_path_list = []
    segment_path = join(condor_dir, "time_segments")
    os.makedirs(segment_path, exist_ok=True)

    for det in time_segments.keys():

        segments = np.array_split(time_segments[det], num_jobs)

        for i, segs in enumerate(segments):

            save_dict = {det: segs}
            filename = join(segment_path, f"seg_{det}_{i:05d}.pkl")
            with open(filename, "wb") as f:
                pickle.dump(save_dict, f)
            time_segments_path_list.append(filename)

    return time_segments_path_list


def create_dag(data_dir, settings_file, time_segments, out_name):
    """
    Create a Condor DAG to (a) download, estimate,
    individual PSDs and (b) merge them into one dataset

    Parameters
    ----------
    data_dir : str
        Path to the directory where the PSD dataset will be stored
    settings_file : str
        Settings : Path to settings file relevant for PSD generation
    time_segments : dict
        contains all time segments used for estimating PSDs
    out_name : str
        path where the resulting ASD dataset should be stored

    Returns
    -------
        Condor DAG
    """

    with open(settings_file, "r") as f:
        settings = yaml.safe_load(f)

    condor_dir = join(data_dir, "tmp", "condor")
    kwargs = {
        "request_cpus": settings["condor"]["num_cpus"],
        "request_memory": settings["condor"]["memory_cpus"],
        "submit": os.path.join(condor_dir, "submit"),
        "error": os.path.join(condor_dir, "error"),
        "output": os.path.join(condor_dir, "output"),
        "log": os.path.join(condor_dir, "logging"),
        "getenv": True,
    }
    # scripts are installed in the env's bin directory
    env_path = os.path.join(settings["condor"]["env_path"], "bin")
    executable = os.path.join(env_path, "dingo_estimate_psds")
    num_jobs = settings["condor"]["num_jobs"]

    time_segment_path_list = split_time_segments(time_segments, condor_dir, num_jobs)

    # DAG ---------------------------------------------------------------------
    dagman = Dagman(
        name="dingo_generate_ASD_dataset_dagman", submit=join(condor_dir, "dag")
    )

    # --- (a) Download and estimate PSDs corresponding to each time segment
    job_list = []
    for seg_path in time_segment_path_list:

        args_dict = {
            "data_dir": data_dir,
            "settings_file": settings_file,
            "time_segments_file": seg_path,
        }
        args_str = create_args_string(args_dict)

        psd_job = Job(
            name="psd_estimation",
            executable=executable,
            dag=dagman,
            arguments=args_str,
            **kwargs,
        )

        job_list.append(psd_job)

    # --- (b) Consolidate dataset
    executable = os.path.join(env_path, "dingo_merge_ASD_datasets")
    args_dict = {
        "data_dir": data_dir,
        "settings_file": settings_file,
        "time_segments_file": join(
            data_dir, "tmp", settings["dataset_settings"]["observing_run"], "psd_time_segments.pkl"
        )
    }
    if out_name is not None:
        args_dict["out_name"] = out_name

    args_str = create_args_string(args_dict)
    consolidate_dataset = Job(
        name="consolidate_dataset",
        executable=executable,
        dag=dagman,
        arguments=args_str,
        **kwargs,
    )
    for job in job_list:
        consolidate_dataset.add_parent(job)

    return dagman


import numpy as np
from gwpy.timeseries import TimeSeries
import pycbc.psd
import math

from dingo.gw.gwutils import get_window


def download_psd(det, time_start, time_psd, window, f_s):
    """
    Download strain data and generate a PSD based on these. Use num_segments of length
    time_segment, starting at GPS time time_start.

    Parameters
    ----------
    det: str
        detector
    time_start: float
        start GPS time for PSD estimation
    time_psd: float = 1024
        time in seconds for strain used for PSD generation
    window: Union(np.ndarray, dict)
        Window used for PSD generation, needs to be the same as used for Fourier
        transform of event strain data.
        Provided as dict, window is generated by window = dingo.gw.gwutils.get_window(
        **window).
    f_s: float
        sampling rate of strain data

    Returns
    -------
    psd: np.array
        array of psd
    """
    # download strain data for psd
    # print("Downloading strain data for PSD estimation.", end=" ")
    time_end = time_start + time_psd
    psd_strain = TimeSeries.fetch_open_data(
        det, time_start, time_end, sample_rate=f_s, cache=True
    )

    # if strain for PSD data contains nan, shift segment for PSD
    if np.max(np.isnan(psd_strain)):
        dt = math.ceil(np.where(np.isnan(psd_strain))[0][-1] / f_s)
        print(
            f"Nan encountered in strain data for PSD estimation for detector {det}. "
            f"Shifting strain segment by {dt} seconds."
        )
        psd_strain = TimeSeries.fetch_open_data(
            det, time_start + dt, time_end + dt, sample_rate=f_s, cache=True
        )
        if np.max(np.isnan(psd_strain)):
            raise ValueError(
                f"Nan encountered in strain data for PSD estimation for detector {det}."
            )

    psd_strain = psd_strain.to_pycbc()

    # generate window
    window = get_window(window)

    # generate PSD from strain data
    psd = pycbc.psd.estimate.welch(
        psd_strain,
        seg_len=len(window),
        seg_stride=len(window),
        window=window,
        avg_method="median",
    )

    return np.array(psd)


def download_raw_data(
    time_event, time_segment, time_psd, time_buffer, detectors, window, f_s
):
    # parse settings
    # time_segment = settings["window"]["T"]  # for now; change later for non-FD data
    # time_psd = settings["time_psd"]
    # time_buffer = settings["time_buffer"]
    # detectors = settings["detectors"]
    # window = settings["window"]

    data = {"strain": {}, "psd": {}}

    for det in detectors:
        data["strain"][det] = TimeSeries.fetch_open_data(
            det,
            time_event + time_buffer - time_segment,
            time_event + time_buffer,
            sample_rate=f_s,
            cache=True,
        )
        data["psd"][det] = download_psd(
            det,
            time_start=time_event + time_buffer - time_psd - time_segment,
            time_psd=time_psd,
            window=window,
            f_s=f_s,
        )

    return data



import pytest
import numpy as np

from dingo.gw.transforms import SelectStandardizeRepackageParameters


def test_SelectStandardizeRepackageParameters():
    standardization_dict = {
        "mean": {"par2": np.random.rand(), "par1": np.random.rand()},
        "std": {"par2": np.random.rand(), "par1": np.random.rand()},
    }
    parameters_dict = {"inference_parameters": ["par1", "par2"]}
    select_standardize_repackage_params = SelectStandardizeRepackageParameters(
        parameters_dict, standardization_dict
    )
    sample = {
        "a": None,
        "b": np.random.rand(100),
        "parameters": {
            "par0": np.random.rand(),
            "par1": np.random.rand(),
            "par2": np.random.rand(),
        },
    }
    sample_out = select_standardize_repackage_params(sample)
    # Check that correct new key has been added.
    assert list(sample.keys()) + ["inference_parameters"] == list(sample_out.keys())
    # check that pre-existing sample elements are not modified
    for k, v in sample.items():
        assert id(v) == id(sample[k])
    # check that correct number of parameters is selected
    assert len(sample_out["inference_parameters"]) == len(
        parameters_dict["inference_parameters"]
    )
    # check that parameter array contains correct elements, in correct order,
    # correctly normalized
    for idx, k in enumerate(parameters_dict["inference_parameters"]):
        m, std = standardization_dict["mean"][k], standardization_dict["std"][k]
        par_in = sample["parameters"][k]
        par_out = sample_out["inference_parameters"][idx]
        # standardization changes dtype to float32
        assert par_out == np.float32((par_in - m) / std)



import argparse
import copy
import textwrap
import numpy as np
from typing import Dict

import yaml

from dingo.gw.noise.asd_dataset import ASDDataset
from dingo.gw.noise.synthetic.asd_parameterization import parameterize_asd_dataset
from dingo.gw.noise.synthetic.asd_sampling import KDE, get_rescaling_params
from dingo.gw.noise.synthetic.utils import reconstruct_psds_from_parameters


def parse_args():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent(
            """\
        Generate a synthetic noise ASD dataset from an existing dataset of real ASDs. ASDs can be parameterized to generate
        smooth PSDs, e.g. to obtain a distribution over ASDs similar to BayesWave ASDs, or we can create a dataset over
        synthetic PSDs to augment the training distribution and enhance robustness. In particular, this allows us to 
        shift a distribution over ASDs from a source to a target observing run, where insufficient data is available.
        """
        ),
    )
    parser.add_argument(
        "--asd_dataset",
        type=str,
        required=True,
        help="Path to existing ASD dataset to be parameterized and re-sampled",
    )
    parser.add_argument(
        "--settings_file",
        type=str,
        required=True,
        help="YAML file containing database settings",
    )
    parser.add_argument(
        "--num_samples",
        type=int,
        default=500,
        help="Number of samples to draw from the parameterized ASDs",
    )
    parser.add_argument(
        "--num_processes",
        type=int,
        default=1,
        help="Number of processes to use in pool for parallel parameterization",
    )
    parser.add_argument(
        "--out_file",
        type=str,
        default="synthetic_asd_dataset.hdf5",
        help="Name of file for storing dataset.",
    )
    parser.add_argument("--verbose", action="store_true")

    return parser.parse_args()


def generate_dataset(real_dataset, settings: Dict, num_samples, num_processes: int, verbose: bool):
    """
    Generate a synthetic ASD dataset from an existing dataset of real ASDs.

    Parameters
    ----------
    real_dataset : ASDDataset
        Existing dataset of real ASDs.
    settings : dict
        Dictionary containing the settings for the parameterization and sampling.
    num_processes : int
        Number of processes to use in pool for parallel parameterization.
    verbose : bool
        Whether to print progress information.

    """
    parameters_dict = parameterize_asd_dataset(
        real_dataset,
        settings["parameterization_settings"],
        num_processes,
        verbose=verbose,
    )

    synthetic_dataset_dict = copy.deepcopy(real_dataset.to_dictionary())
    synthetic_dataset_dict["settings"]["parameterization_settings"] = settings[
        "parameterization_settings"
    ]

    sampling_settings = settings.get("sampling_settings")
    if sampling_settings:
        kde = KDE(parameters_dict, sampling_settings)
        kde.fit()
        rescaling_params = None
        if "rescaling_asd_paths" in sampling_settings:
            rescaling_params = get_rescaling_params(
                sampling_settings["rescaling_asd_paths"],
                settings["parameterization_settings"],
            )
        parameters_dict = kde.sample(num_samples, rescaling_params)
        synthetic_dataset_dict["settings"]["sampling_settings"] = settings["sampling_settings"]

    asds_dict = {}
    for det, params in parameters_dict.items():
        psds = reconstruct_psds_from_parameters(
            params,
            real_dataset.domain,
            settings["parameterization_settings"],
        )

        asds_dict[det] = np.sqrt(
            psds[:, real_dataset.domain.min_idx : real_dataset.domain.max_idx + 1]
        )

    synthetic_dataset_dict["asd_parameterizations"] = parameters_dict
    synthetic_dataset_dict["asds"] = asds_dict

    return ASDDataset(
        dictionary=synthetic_dataset_dict
    )


def main():
    args = parse_args()

    # Load settings
    with open(args.settings_file, "r") as f:
        settings = yaml.safe_load(f)

    real_dataset = ASDDataset(file_name=args.asd_dataset)
    synthetic_dataset = generate_dataset(
        real_dataset, settings, args.num_samples, args.num_processes, args.verbose
    )

    synthetic_dataset.to_file(args.out_file)


if __name__ == "__main__":
    main()



import os
import numpy as np
import yaml
import argparse
import textwrap


from threadpoolctl import threadpool_limits

from dingo.core.nn.nsf import autocomplete_model_kwargs_nsf
from dingo.gw.training.train_builders import (
    build_dataset,
    build_train_and_test_loaders,
    set_train_transforms,
    build_svd_for_embedding_network,
)
from dingo.core.models.posterior_model import PosteriorModel
from dingo.core.utils.trainutils import RuntimeLimits
from dingo.core.utils import (
    set_requires_grad_flag,
    get_number_of_model_parameters,
    build_train_and_test_loaders,
)
from dingo.gw.dataset import WaveformDataset


def prepare_training_new(train_settings: dict, train_dir: str, local_settings: dict):
    """
    Based on a settings dictionary, initialize a WaveformDataset and PosteriorModel.

    For model type 'nsf+embedding' (the only acceptable type at this point) this also
    initializes the embedding network projection stage with SVD V matrices based on
    clean detector waveforms.

    Parameters
    ----------
    train_settings : dict
        Settings which ultimately come from train_settings.yaml file.
    train_dir : str
        This is only used to save diagnostics from the SVD.
    local_settings : dict
        Local settings (e.g., num_workers, device)

    Returns
    -------
    (WaveformDataset, PosteriorModel)
    """

    wfd = build_dataset(train_settings["data"])  # No transforms yet
    initial_weights = {}

    # This is the only case that exists so far, but we leave it open to develop new
    # model types.
    if train_settings["model"]["type"] == "nsf+embedding":

        # First, build the SVD for seeding the embedding network.
        print("\nBuilding SVD for initialization of embedding network.")
        initial_weights["V_rb_list"] = build_svd_for_embedding_network(
            wfd,
            train_settings["data"],
            train_settings["training"]["stage_0"]["asd_dataset_path"],
            num_workers=local_settings["num_workers"],
            batch_size=train_settings["training"]["stage_0"]["batch_size"],
            out_dir=train_dir,
            **train_settings["model"]["embedding_net_kwargs"]["svd"],
        )

        # Now set the transforms for training. We need to do this here so that we can (a)
        # get the data dimensions to configure the network, and (b) save the
        # parameter standardization dict in the PosteriorModel. In principle, (a) could
        # be done without generating data (by careful calculation) and (b) could also
        # be done outside the transform setup. But for now, this is convenient. The
        # transforms will be reset later by initialize_stage().

        set_train_transforms(
            wfd,
            train_settings["data"],
            train_settings["training"]["stage_0"]["asd_dataset_path"],
        )

        # This modifies the model settings in-place.
        autocomplete_model_kwargs_nsf(train_settings["model"], wfd[0])
        full_settings = {
            "dataset_settings": wfd.settings,
            "train_settings": train_settings,
        }

    else:
        raise ValueError('Model type must be "nsf+embedding".')

    print("\nInitializing new posterior model.")
    print("Complete settings:")
    print(yaml.dump(full_settings, default_flow_style=False, sort_keys=False))

    pm = PosteriorModel(
        metadata=full_settings,
        initial_weights=initial_weights,
        device=local_settings["device"],
    )

    if local_settings.get("wandb", False):
        try:
            import wandb
            wandb.init(
                config=full_settings,
                dir=train_dir,
                **local_settings["wandb"],
            )
        except ImportError:
            print("WandB is enabled but not installed.")

    return pm, wfd


def prepare_training_resume(checkpoint_name, local_settings, train_dir):
    """
    Loads a PosteriorModel from a checkpoint, as well as the corresponding
    WaveformDataset, in order to continue training. It initializes the saved optimizer
    and scheduler from the checkpoint.

    Parameters
    ----------
    checkpoint_name : str
        File name containing the checkpoint (.pt format).
    device : str
        'cuda' or 'cpu'

    Returns
    -------
    (PosteriorModel, WaveformDataset)
    """

    pm = PosteriorModel(model_filename=checkpoint_name, device=local_settings["device"])
    wfd = build_dataset(pm.metadata["train_settings"]["data"])

    if local_settings.get("wandb", False):
        try:
            import wandb
            wandb.init(
                resume="must",
                dir=train_dir,
                **local_settings["wandb"],
            )
        except ImportError:
            print("WandB is enabled but not installed.")

    return pm, wfd


def initialize_stage(pm, wfd, stage, num_workers, resume=False):
    """
    Initializes training based on PosteriorModel metadata and current stage:
        * Builds transforms (based on noise settings for current stage);
        * Builds DataLoaders;
        * At the beginning of a stage (i.e., if not resuming mid-stage), initializes
        a new optimizer and scheduler;
        * Freezes / unfreezes SVD layer of embedding network

    Parameters
    ----------
    pm : PosteriorModel
    wfd : WaveformDataset
    stage : dict
        Settings specific to current stage of training
    num_workers : int
    resume : bool
        Whether training is resuming mid-stage. This controls whether the optimizer and
        scheduler should be re-initialized based on contents of stage dict.

    Returns
    -------
    (train_loader, test_loader)
    """

    train_settings = pm.metadata["train_settings"]

    # Rebuild transforms based on possibly different noise.
    set_train_transforms(wfd, train_settings["data"], stage["asd_dataset_path"])

    # Allows for changes in batch size between stages.
    train_loader, test_loader = build_train_and_test_loaders(
        wfd,
        train_settings["data"]["train_fraction"],
        stage["batch_size"],
        num_workers,
    )

    if not resume:
        # New optimizer and scheduler. If we are resuming, these should have been
        # loaded from the checkpoint.
        print("Initializing new optimizer and scheduler.")
        pm.optimizer_kwargs = stage["optimizer"]
        pm.scheduler_kwargs = stage["scheduler"]
        pm.initialize_optimizer_and_scheduler()

    # Freeze/unfreeze RB layer if necessary
    if "freeze_rb_layer" in stage:
        if stage["freeze_rb_layer"]:
            set_requires_grad_flag(
                pm.model, name_contains="layers_rb", requires_grad=False
            )
        else:
            set_requires_grad_flag(
                pm.model, name_contains="layers_rb", requires_grad=True
            )
    n_grad = get_number_of_model_parameters(pm.model, (True,))
    n_nograd = get_number_of_model_parameters(pm.model, (False,))
    print(f"Fixed parameters: {n_nograd}\nLearnable parameters: {n_grad}\n")

    return train_loader, test_loader


def train_stages(pm, wfd, train_dir, local_settings):
    """
    Train the network, iterating through the sequence of stages. Stages can change
    certain settings such as the noise characteristics, optimizer, and scheduler settings.

    Parameters
    ----------
    pm : PosteriorModel
    wfd : WaveformDataset
    train_dir : str
        Directory for saving checkpoints and train history.
    local_settings : dict

    Returns
    -------
    bool
        True if all stages are complete
        False otherwise
    """

    train_settings = pm.metadata["train_settings"]
    runtime_limits = RuntimeLimits(
        epoch_start=pm.epoch, **local_settings["runtime_limits"]
    )

    # Extract list of stages from settings dict
    stages = []
    num_stages = 0
    while True:
        try:
            stages.append(train_settings["training"][f"stage_{num_stages}"])
            num_stages += 1
        except KeyError:
            break
    end_epochs = list(np.cumsum([stage["epochs"] for stage in stages]))

    num_starting_stage = np.searchsorted(end_epochs, pm.epoch + 1)
    for n in range(num_starting_stage, num_stages):
        stage = stages[n]

        if pm.epoch == end_epochs[n] - stage["epochs"]:
            print(f"\nBeginning training stage {n}. Settings:")
            print(yaml.dump(stage, default_flow_style=False, sort_keys=False))
            train_loader, test_loader = initialize_stage(
                pm, wfd, stage, local_settings["num_workers"], resume=False
            )
        else:
            print(f"\nResuming training in stage {n}. Settings:")
            print(yaml.dump(stage, default_flow_style=False, sort_keys=False))
            train_loader, test_loader = initialize_stage(
                pm, wfd, stage, local_settings["num_workers"], resume=True
            )

        runtime_limits.max_epochs_total = end_epochs[n]
        pm.train(
            train_loader,
            test_loader,
            train_dir=train_dir,
            runtime_limits=runtime_limits,
            checkpoint_epochs=local_settings["checkpoint_epochs"],
            use_wandb=local_settings.get("wandb", False),
            test_only=local_settings.get("test_only", False),
        )
        # if test_only, model should not be saved, and run is complete
        if local_settings.get("test_only", False):
            return True

        if pm.epoch == end_epochs[n]:
            save_file = os.path.join(train_dir, f"model_stage_{n}.pt")
            print(f"Training stage complete. Saving to {save_file}.")
            pm.save_model(save_file, save_training_info=True)
        if runtime_limits.local_limits_exceeded(pm.epoch):
            print("Local runtime limits reached. Ending program.")
            break

    if pm.epoch == end_epochs[-1]:
        return True
    else:
        return False


def parse_args():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent(
            """\
        Train a neural network for gravitational-wave single-event inference.
        
        This program can be called in one of two ways:
            a) with a settings file. This will create a new network based on the 
            contents of the settings file.
            b) with a checkpoint file. This will resume training from the checkpoint.
        """
        ),
    )
    parser.add_argument(
        "--settings_file",
        type=str,
        help="YAML file containing training settings.",
    )
    parser.add_argument(
        "--train_dir", required=True, help="Directory for Dingo training output."
    )
    parser.add_argument(
        "--checkpoint",
        type=str,
        help="Checkpoint file from which to resume training.",
    )
    args = parser.parse_args()

    # The settings file and checkpoint are mutually exclusive.
    if args.checkpoint is None and args.settings_file is None:
        parser.error("Must specify either a checkpoint file or a settings file.")
    if args.checkpoint is not None and args.settings_file is not None:
        parser.error("Cannot specify both a checkpoint file and a settings file.")

    return args


def train_local():

    args = parse_args()

    os.makedirs(args.train_dir, exist_ok=True)

    if args.settings_file is not None:
        print("Beginning new training run.")
        with open(args.settings_file, "r") as fp:
            train_settings = yaml.safe_load(fp)

        # Extract the local settings from train settings file, save it separately. This
        # file can later be modified, and the settings take effect immediately upon
        # resuming.

        local_settings = train_settings.pop("local")
        with open(os.path.join(args.train_dir, "local_settings.yaml"), "w") as f:
            if local_settings.get("wandb", False) and "id" not in local_settings["wandb"].keys():
                try:
                    import wandb
                    local_settings["wandb"]["id"] = wandb.util.generate_id()
                except ImportError:
                    print("wandb not installed, cannot generate run id.")
            yaml.dump(local_settings, f, default_flow_style=False, sort_keys=False)

        pm, wfd = prepare_training_new(train_settings, args.train_dir, local_settings)

    else:
        print("Resuming training run.")
        with open(os.path.join(args.train_dir, "local_settings.yaml"), "r") as f:
            local_settings = yaml.safe_load(f)
        pm, wfd = prepare_training_resume(
            args.checkpoint, local_settings, args.train_dir
        )

    with threadpool_limits(limits=1, user_api="blas"):
        complete = train_stages(pm, wfd, args.train_dir, local_settings)

    if complete:
        print("All training stages complete.")
    else:
        print("Program terminated due to runtime limit.")



import os

from bilby_pipe.job_creation.nodes import GenerationNode as BilbyGenerationNode

from dingo.pipe.utils import _strip_unwanted_submission_keys


class GenerationNode(BilbyGenerationNode):

    def __init__(self, inputs, importance_sampling=False, **kwargs):
        self.importance_sampling = importance_sampling
        super().__init__(inputs, **kwargs)

        if self.inputs.simple_submission:
            _strip_unwanted_submission_keys(self.job)

    @property
    def executable(self):
        return self._get_executable_path("dingo_pipe_generation")

    def setup_arguments(self, **kwargs):
        super().setup_arguments(**kwargs)
        if self.importance_sampling:
            self.arguments.add_flag("importance-sampling-generation")

    @property
    def job_name(self):
        flag = "_IS" if self.importance_sampling else ""
        return super().job_name + flag

    @property
    def event_data_file(self):
        return os.path.join(
            self.inputs.data_directory, "_".join([self.label, "event_data.hdf5"])
        )



import pytest
import numpy as np
import pandas as pd
from dingo.gw.conversion import cartesian_spins, pe_spins, change_spin_conversion_phase


@pytest.fixture
def param_setup():
    f_ref = 20
    params = {
        "mass_ratio": 0.9834472418378746,
        "chirp_mass": 56.75982559976776,
        "luminosity_distance": 1000.0,
        "theta_jn": 1.913533991012538,
        "phase": 0.27093919407580236,
        "a_1": 0.23877963026418533,
        "a_2": 0.04520077119798103,
        "tilt_1": 1.5215996866276775,
        "tilt_2": 1.323159785592912,
        "phi_12": 0.8296989755334262,
        "phi_jl": 3.5628496275567447,
        "geocent_time": 0.0,
    }
    return f_ref, params


def test_spin_conversions(param_setup):
    f_ref, params = param_setup
    params_cart = cartesian_spins(params, f_ref)
    params_pe = pe_spins(params_cart, f_ref)
    params_pe_wrong = pe_spins(params_cart, f_ref + 10)

    assert params.keys() == params_pe.keys()
    assert np.sum([np.abs(v - params_pe[k]) for k, v in params.items()]) < 1e-10
    assert np.sum([np.abs(v - params_pe_wrong[k]) for k, v in params.items()]) > 1e-3


def test_change_of_spin_conversion_phase(param_setup):
    f_ref, params = param_setup
    params_pd = pd.DataFrame({k: np.array([v, v / 1.2]) for k, v in params.items()})

    # 1) check that only theta_jn and phi_jl are impacted
    params_pd_new = change_spin_conversion_phase(params_pd, f_ref, None, 0)
    diff = {
        k: np.sum(np.abs(params_pd_new[k] - params_pd[k])) for k in params_pd.keys()
    }
    for k in ["theta_jn", "phi_jl"]:
        assert diff[k] > 1e-3
    for k in [kk for kk in diff.keys() if kk not in ["theta_jn", "phi_jl"]]:
        assert diff[k] < 1e-10

    # 2) check that changing phases back and forth is consistent
    params_pd_new = params_pd.copy()
    params_pd_new = change_spin_conversion_phase(params_pd_new, f_ref, None, 0)
    params_pd_new = change_spin_conversion_phase(params_pd_new, f_ref, 0, 1)
    params_pd_new = change_spin_conversion_phase(params_pd_new, f_ref, 1, None)
    diff = {
        k: np.sum(np.abs(params_pd_new[k] - params_pd[k])) for k in params_pd.keys()
    }
    for v in diff.values():
        assert v < 1e-10



"""Implementation of embedding networks."""

from typing import Tuple, Callable, Union, List
import torch
import numpy as np
import torch.nn as nn
from torch.nn import functional as F
from glasflow.nflows.nn.nets.resnet import ResidualBlock
from dingo.core.utils import torchutils


class LinearProjectionRB(nn.Module):
    """
    A compression layer that reduces the input dimensionality via projection
    onto a reduced basis. The input data is of shape (batch_size, num_blocks,
    num_channels, num_bins). Each of the num_blocks blocks (for GW use case:
    block=detector) is treated independently.

    A single block consists of 1D data with num_bins bins (e.g. GW use case:
    num_bins=number of frequency bins). It has num_channels>=2 different
    channels, channel 0 and 1 store the real and imaginary part of the
    signal. Channels with index >=2 are used for auxiliary signals (such as
    PSD for GW use case).

    This layer compresses the complex signal in channels 0 and 1 to n_rb
    reduced-basis (rb) components. This is achieved by initializing the
    weights of this layer with the rb matrix V, such that the (2*n_rb)
    dimensional output of each block is the concatenation of the real and
    imaginary part of the reduced basis projection of the complex signal in
    channel 0 and 1. The projection of the auxiliary channels with index >=2
    onto these components is initialized with 0.

    Module specs
    --------
        input dimension:    (batch_size, num_blocks, num_channels, num_bins)
        output dimension:   (batch_size, 2 * n_rb * num_blocks)
    """

    def __init__(
        self,
        input_dims: List[int],
        n_rb: int,
        V_rb_list: Union[Tuple, None],
    ):
        """
        Parameters
        ----------
        input_dims : list
            dimensions of input batch, omitting batch dimension
            input_dims = [num_blocks, num_channels, num_bins]
        n_rb : int
            number of reduced basis elements used for projection
            the output dimension of the layer is 2 * n_rb * num_blocks
        V_rb_list : tuple of np.arrays, or None
            tuple with V matrices of the reduced basis SVD projection,
            convention for SVD matrix decomposition: U @ s @ V^h;
            if None, layer is not initialized with reduced basis projection,
            this is useful when loading a saved model
        """

        super(LinearProjectionRB, self).__init__()

        self.input_dims = input_dims
        self.num_blocks, self.num_channels, self.num_bins = self.input_dims
        self.n_rb = n_rb

        # define a linear projection layer for each block
        layers = []
        for _ in range(self.num_blocks):
            layers.append(nn.Linear(self.num_bins * self.num_channels, self.n_rb * 2))
        self.layers_rb = nn.ModuleList(layers)

        # initialize layers with reduced basis
        if V_rb_list is not None:
            if type(V_rb_list[0]) == str:
                V_rb_list = [np.load(el) for el in V_rb_list]
            self.test_dimensions(V_rb_list)
            self.init_layers(V_rb_list)

    @property
    def input_dim(self):
        return self.num_bins * self.num_channels * self.num_blocks

    @property
    def output_dim(self):
        return 2 * self.n_rb * self.num_blocks

    def test_dimensions(self, V_rb_list):
        """Test if input dimensions to this layer are consistent with each
        other, and the reduced basis matrices V."""
        if self.num_channels < 2:
            raise ValueError(
                "Number of channels needs to be at least 2, for real and "
                "imaginary parts."
            )
        if len(V_rb_list) != self.num_blocks:
            raise ValueError(
                "There must be exactly one reduced basis matrix V for each " "block."
            )
        for V in V_rb_list:
            if not isinstance(V, np.ndarray) or len(V.shape) != 2:
                raise ValueError(
                    "Reduced basis matrix V must be a numpy array with 2 axes."
                )
            if V.shape[0] != self.num_bins:
                raise ValueError(
                    "Number of input bins and number of rows in rb matrix V "
                    "need to match."
                )
            if V.shape[1] < self.n_rb:
                raise ValueError(
                    "More reduced basis elements requested than available."
                )

    def init_layers(self, V_rb_list):
        """
        Loop through layers and initialize them individually with the
        corresponding rb projection. V_rb_list is a list that contains the rb
        matrix V for each block. Each matrix V in V_rb_list is represented
        with a numpy array of shape (self.num_bins, num_el), where
        num_el >= self.n_rb.
        """
        n = self.n_rb
        k = self.num_bins
        for ind, layer in enumerate(self.layers_rb):
            V = V_rb_list[ind]

            # truncate V to n_rb basis elements
            V = V[:, :n]
            V_real, V_imag = (
                torch.from_numpy(V.real).float(),
                torch.from_numpy(V.imag).float(),
            )

            # initialize all weights and biases with zero
            layer.weight.data = torch.zeros_like(layer.weight.data)
            layer.bias.data = torch.zeros_like(layer.bias.data)

            # load matrix V into weights
            layer.weight.data[:n, :k] = torch.transpose(V_real, 1, 0)
            layer.weight.data[n:, :k] = torch.transpose(V_imag, 1, 0)
            layer.weight.data[:n, k : 2 * k] = -torch.transpose(V_imag, 1, 0)
            layer.weight.data[n:, k : 2 * k] = torch.transpose(V_real, 1, 0)

    def forward(self, x):
        if x.shape[1:] != (self.num_blocks, self.num_channels, self.num_bins):
            raise ValueError(
                f"Invalid shape for projection layer. "
                f"Expected {(self.num_blocks, self.num_channels, self.num_bins)}, "
                f"got {tuple(x.shape[1:])}."
            )
        out = []
        for ind in range(self.num_blocks):
            out.append(self.layers_rb[ind](x[:, ind, ...].flatten(start_dim=1)))
        x = torch.cat(out, dim=1)
        return x


class DenseResidualNet(nn.Module):
    """
    A nn.Module consisting of a sequence of dense residual blocks. This is
    used to embed high dimensional input to a compressed output. Linear
    resizing layers are used for resizing the input and output to match the
    first and last hidden dimension, respectively.

    Module specs
    --------
        input dimension:    (batch_size, input_dim)
        output dimension:   (batch_size, output_dim)
    """

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        hidden_dims: Tuple,
        activation: Callable = F.elu,
        dropout: float = 0.0,
        batch_norm: bool = True,
    ):
        """
        Parameters
        ----------
        input_dim : int
            dimension of the input to this module
        output_dim : int
            output dimension of this module
        hidden_dims : tuple
            tuple with dimensions of hidden layers of this module
        activation: callable
            activation function used in residual blocks
        dropout: float
            dropout probability for residual blocks used for reqularization
        batch_norm: bool
            flag that specifies whether to use batch normalization
        """

        super(DenseResidualNet, self).__init__()
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.hidden_dims = hidden_dims
        self.num_res_blocks = len(self.hidden_dims)

        self.initial_layer = nn.Linear(self.input_dim, hidden_dims[0])
        self.blocks = nn.ModuleList(
            [
                ResidualBlock(
                    features=self.hidden_dims[n],
                    context_features=None,
                    activation=activation,
                    dropout_probability=dropout,
                    use_batch_norm=batch_norm,
                )
                for n in range(self.num_res_blocks)
            ]
        )
        self.resize_layers = nn.ModuleList(
            [
                nn.Linear(self.hidden_dims[n - 1], self.hidden_dims[n])
                if self.hidden_dims[n - 1] != self.hidden_dims[n]
                else nn.Identity()
                for n in range(1, self.num_res_blocks)
            ]
            + [nn.Linear(self.hidden_dims[-1], self.output_dim)]
        )

    def forward(self, x):
        x = self.initial_layer(x)
        for block, resize_layer in zip(self.blocks, self.resize_layers):
            x = block(x, context=None)
            x = resize_layer(x)
        return x


class ModuleMerger(nn.Module):
    """
    This is a wrapper used to process multiple different kinds of context
    information collected in x = (x_0, x_1, ...). For each kind of context
    information x_i, an individual embedding network is provided in
    enets = (enet_0, enet_1, ...). The embedded output of the forward method
    is the concatenation of the individual embeddings enet_i(x_i).

    In the GW use case, this wrapper can be used to embed the
    high-dimensional signal input into a lower dimensional feature vector
    with a large embedding network, while applying an identity embedding to
    the time shifts.

    Module specs
    --------
        input dimension:    (batch_size, ...), (batch_size, ...), ...
        output dimension:   (batch_size, ?)
    """

    def __init__(
        self,
        module_list: Tuple,
    ):
        """
        Parameters
        ----------
        module_list : tuple
            nn.Modules for embedding networks,
            use torch.nn.Identity for identity mappings
        """
        super(ModuleMerger, self).__init__()
        self.enets = nn.ModuleList(module_list)

    def forward(self, *x):
        if len(x) != len(self.enets):
            raise ValueError("Invalid number of input tensors provided.")
        x = [module(xi) for module, xi in zip(self.enets, x)]
        return torch.cat(x, axis=1)


def create_enet_with_projection_layer_and_dense_resnet(
    input_dims: List[int],
    # n_rb: int,
    V_rb_list: Union[Tuple, None],
    output_dim: int,
    hidden_dims: Tuple,
    svd: dict,
    activation: str = "elu",
    dropout: float = 0.0,
    batch_norm: bool = True,
    added_context: bool = False,
):
    """
    Builder function for 2-stage embedding network for 1D data with multiple
    blocks and channels. Module 1 is a linear layer initialized as the
    projection of the complex signal onto reduced basis components via the
    LinearProjectionRB, where the blocks are kept separate. See docstring
    of LinearProjectionRB for details. Module 2 is a sequence of dense residual
    layers, that is used to further reduce the dimensionality.

    The projection requires the complex signal to be represented via the real
    part in channel 0 and the imaginary part in channel 1. Auxiliary signals
    may be contained in channels with indices => 2. In GW use case a block
    corresponds to a detector and channel 2 is used for ASD information.

    If added_context = True, the 2-stage embedding network described above is
    merged with an identity mapping via ModuleMerger. Then, the expected input
    is not x with x.shape = (batch_size, num_blocks, num_channels, num_bins),
    but rather the tuple *(x, z), where z is additional context information. The
    output of the full module is then the concatenation of enet(x) and z. In
    GW use case, this is used to concatenate the applied time shifts z to the
    embedded feature vector of the strain data enet(x).

    Module specs
    --------
    For added_context == False:
        input dimension:    (batch_size, num_blocks, num_channels, num_bins)
        output dimension:   (batch_size, output_dim)
    For added_context == True:
        input dimension:    (batch_size, num_blocks, num_channels, num_bins),
                            (batch_size, N)
        output dimension:   (batch_size, output_dim + N)

    :param input_dims:  list
        dimensions of input batch, omitting batch dimension
        input_dims = (num_blocks, num_channels, num_bins)
    :param n_rb: int
        number of reduced basis elements used for projection
        the output dimension of the layer is 2 * n_rb * num_blocks
    :param V_rb_list: tuple of np.arrays, or None
        tuple with V matrices of the reduced basis SVD projection,
        convention for SVD matrix decomposition: U @ s @ V^h;
        if None, layer is not initialized with reduced basis projection,
        this is useful when loading a saved model
    :param output_dim: int
        output dimension of the full module
    :param hidden_dims: tuple
        tuple with dimensions of hidden layers of module 2
    :param activation: str
        str that specifies activation function used in residual blocks
    :param dropout: float
        dropout probability for residual blocks used for reqularization
    :param batch_norm: bool
        flag that specifies whether to use batch normalization
    :param added_context: bool
        if set to True, additional context z is concatenated to the embedded
        feature vector enet(x); note that in this case, the expected input is
        a tuple with 2 elements, input = (x, z) rather than just the tensor x.
    :return: nn.Module
    """
    activation_fn = torchutils.get_activation_function_from_string(activation)
    module_1 = LinearProjectionRB(input_dims, svd["size"], V_rb_list)
    module_2 = DenseResidualNet(
        input_dim=module_1.output_dim,
        output_dim=output_dim,
        hidden_dims=hidden_dims,
        activation=activation_fn,
        dropout=dropout,
        batch_norm=batch_norm,
    )
    enet = nn.Sequential(module_1, module_2)

    if not added_context:
        return enet
    else:
        return ModuleMerger((enet, nn.Identity()))


if __name__ == "__main__":
    pass



"""
This submodule contains tools for density estimation from samples.
This is required for instance to recover the posterior density from GNPE samples,
since the density is intractable with GNPE.
"""
from .unconditional_density_estimation import train_unconditional_density_estimator
from .interpolation import (
    interpolated_sample_and_log_prob_multi,
    interpolated_log_prob_multi,
)
from .nde_settings import get_default_nde_settings_3d



import torch
import argparse

import yaml


def parse_args():

    parser = argparse.ArgumentParser()
    parser.add_argument("--checkpoint", type=str, required=True)
    parser.add_argument("--out_file", type=str, required=True)
    return parser.parse_args()


def main():
    args = parse_args()

    # Typically training is done on the GPU, so the model could be saved on a GPU
    # device. Since this routine may be run on a CPU machine, allow for a remap of the
    # torch tensors.
    if torch.cuda.is_available():
        d = torch.load(args.checkpoint)
    else:
        d = torch.load(args.checkpoint, map_location=torch.device("cpu"))

    data_settings = d["metadata"]["train_settings"]["data"]

    if "gnpe_time_shifts" in data_settings:
        if data_settings["gnpe_time_shifts"]["exact_equiv"]:

            # Rename context variables "L1_time_proxy" to "L1_time_proxy_relative", etc.
            for i, p in enumerate(data_settings["context_parameters"]):
                if p.endswith('_time_proxy'):
                    data_settings["context_parameters"][i] = p + '_relative'
                    data_settings["standardization"]["mean"][p + '_relative'] = \
                        data_settings["standardization"]["mean"].pop(p)
                    data_settings["standardization"]["std"][p + '_relative'] = \
                        data_settings["standardization"]["std"].pop(p)

    print("New data_settings:")
    print(yaml.dump(data_settings, default_flow_style=False, sort_keys=False))

    torch.save(d, args.out_file)


if __name__ == "__main__":
    main()


import os
from os.path import join
import yaml


def resubmit_condor_job(train_dir, train_settings, epoch):
    """
    TODO: documentation
    :param train_dir:
    :param train_settings:
    :param epoch:
    :return:
    """
    if 'condor_settings' in train_settings:
        print('Copying log files')
        copy_logfiles(train_dir, epoch=epoch)

        if epoch >= train_settings['train_settings']['runtime_limits'][
            'max_epochs_total']:
            print('Training complete, job will not be resubmitted')
        else:
            print('Training incomplete, resubmitting job.')
            create_submission_file_and_submit_job(train_dir)


def create_submission_file_and_submit_job(train_dir,
                                          filename='submission_file.sub'):
    """
    TODO: documentation
    :param train_dir:
    :param filename:
    :return:
    """
    create_submission_file(train_dir, filename)
    with open(join(train_dir, 'train_settings.yaml'), 'r') as fp:
        bid = yaml.safe_load(fp)['condor_settings']['bid']
    os.system(f'condor_submit_bid {bid} {join(train_dir, filename)}')


def create_submission_file(train_dir, filename='submission_file.sub'):
    """
    TODO: documentation
    :param train_dir:
    :param filename:
    :return:
    """
    with open(join(train_dir, 'train_settings.yaml'), 'r') as fp:
        d = yaml.safe_load(fp)['condor_settings']
    lines = []
    lines.append(f'executable = {d["python"]}\n')
    lines.append(f'request_cpus = {d["num_cpus"]}\n')
    lines.append(f'request_memory = {d["memory_cpus"]}\n')
    lines.append(f'request_gpus = {d["num_gpus"]}\n')
    lines.append(f'requirements = TARGET.CUDAGlobalMemoryMb > '
                 f'{d["memory_gpus"]}\n\n')

    lines.append(f'arguments = {d["train_script"]} --train_dir {train_dir}\n')
    lines.append(f'error = {join(train_dir, "info.err")}\n')
    lines.append(f'output = {join(train_dir, "info.out")}\n')
    lines.append(f'log = {join(train_dir, "info.log")}\n')
    lines.append('queue')

    with open(join(train_dir, filename), 'w') as f:
        for line in lines:
            f.write(line)


def copyfile(src, dst):
    os.system('cp -p %s %s' % (src, dst))


def copy_logfiles(log_dir, epoch, name='info', suffixes=('.err','.log','.out')):
    for suffix in suffixes:
        src = join(log_dir, name + suffix)
        dest = join(log_dir, name + '_{:03d}'.format(epoch) + suffix)
        try:
            copyfile(src, dest)
        except:
            print('Could not copy ' + src)


if __name__ == '__main__':
    train_dir = '/Users/mdax/Documents/dingo/devel/dingo-devel/tutorials/02_gwpe/train_dir/'
    create_submission_file(train_dir)

    # epoch = pm.epoch - 1
    # if args.checkpoint_frequency is None:
    #     # save model and copy logfiles
    #     print('Saving model to model_{:03d}.pt.'.format(epoch))
    #     pm.save_model(filename='model_{:03d}.pt'.format(epoch))
    # print('Copying logfiles.')
    # copy_logfiles(pm.history_dir, epoch=epoch)
    #
    # # If training is not finished, resubmit the submission file
    # if epoch >= args.epochs_total:
    #     print('Training complete, job will not be resubmitted')
    # elif args.restart:
    #     create_submission_file(args.history_dir, name='submission_file_x.sub')
    #     resubmit_condor_job(
    #         submission_file=join(args.history_dir, 'submission_file_x.sub'),
    #         bid=5)



#
#  Adapted from bilby_pipe. In particular, uses the bilby_pipe data generation code.
#
import os

from bilby_pipe.input import Input
from bilby_pipe.main import MainInput as BilbyMainInput
from bilby_pipe.utils import (
    parse_args,
    get_command_line_arguments,
    logger,
    convert_string_to_dict,
)

from .dag_creator import generate_dag
from .parser import create_parser

from dingo.gw.domains import build_domain_from_model_metadata
from dingo.core.models import PosteriorModel

logger.name = "dingo_pipe"


def fill_in_arguments_from_model(args):
    # FIXME: It would be better if we did not have to load an entire model just to
    #  gain access to the metadata. Store a copy of metadata separately?
    logger.info(f"Loading dingo model from {args.model} in order to access settings.")
    model = PosteriorModel(args.model, device="cpu", load_training_info=False)
    model_metadata = model.metadata

    domain = build_domain_from_model_metadata(model_metadata)

    data_settings = model_metadata["train_settings"]["data"]

    model_args = {
        "duration": domain.duration,
        "minimum_frequency": domain.f_min,
        "maximum_frequency": domain.f_max,
        "detectors": data_settings["detectors"],
        "sampling_frequency": data_settings["window"]["f_s"],
        "tukey_roll_off": data_settings["window"]["roll_off"],
        "waveform_approximant": model_metadata["dataset_settings"][
            "waveform_generator"
        ][
            "approximant"
        ],  # TODO: Update approximant in IS
    }

    changed_args = {}
    for k, v in model_args.items():
        args_v = getattr(args, k)
        if args_v is not None:

            # Convert type from str to enable comparison.
            try:
                if isinstance(v, float):
                    args_v = float(args_v)
                elif isinstance(v, int):
                    args_v = int(args_v)
            except ValueError:
                pass

            if args_v != v:
                logger.warning(
                    f"Argument {k} provided to dingo_pipe as {args_v} "
                    f"does not match value {v} in model file. Using model value for "
                    f"inference, and will attempt to change this during importance "
                    f"sampling."
                )
                changed_args[k] = args_v

        setattr(args, k, v)

    # TODO: Also check consistency between model and init_model settings.

    # Updates that are explicitly provided take priority.
    if args.importance_sampling_updates is None:
        importance_sampling_updates = {}
    else:
        importance_sampling_updates = convert_string_to_dict(
            args.importance_sampling_updates
        )
        importance_sampling_updates = {
            k.replace("-", "_"): v for k, v in importance_sampling_updates.items()
        }
    return {**changed_args, **importance_sampling_updates}, model_args


class MainInput(BilbyMainInput):
    def __init__(self, args, unknown_args, importance_sampling_updates):

        # Settings added for dingo.

        self.model = args.model
        self.model_init = args.model_init
        self.num_gnpe_iterations = args.num_gnpe_iterations
        self.importance_sampling_updates = importance_sampling_updates

        Input.__init__(self, args, unknown_args, print_msg=False)

        self.known_args = args
        self.unknown_args = unknown_args
        self.ini = args.ini
        self.submit = args.submit
        self.condor_job_priority = args.condor_job_priority
        self.create_summary = args.create_summary

        self.outdir = args.outdir
        self.label = args.label
        self.log_directory = args.log_directory
        self.accounting = args.accounting
        self.accounting_user = args.accounting_user
        # self.sampler = args.sampler
        self.detectors = args.detectors
        self.coherence_test = (
            False  # dingo mod: Cannot use different sets of detectors.
        )
        self.n_parallel = args.n_parallel
        # self.transfer_files = args.transfer_files
        self.osg = args.osg
        self.desired_sites = args.desired_sites
        # self.analysis_executable = args.analysis_executable
        # self.analysis_executable_parser = args.analysis_executable_parser
        self.result_format = "hdf5"
        self.final_result = args.final_result
        self.final_result_nsamples = args.final_result_nsamples

        self.webdir = args.webdir
        self.email = args.email
        self.notification = args.notification
        self.queue = args.queue
        self.existing_dir = args.existing_dir

        self.scheduler = args.scheduler
        self.scheduler_args = args.scheduler_args
        self.scheduler_module = args.scheduler_module
        self.scheduler_env = args.scheduler_env
        self.scheduler_analysis_time = args.scheduler_analysis_time

        # self.waveform_approximant = args.waveform_approximant
        #
        # self.time_reference = args.time_reference
        self.time_reference = "geocent"
        # self.reference_frame = args.reference_frame
        # self.likelihood_type = args.likelihood_type
        self.duration = args.duration
        # self.phase_marginalization = args.phase_marginalization
        # self.prior_file = args.prior_file
        self.prior_dict = args.prior_dict
        self.default_prior = "PriorDict"
        self.minimum_frequency = args.minimum_frequency
        # self.enforce_signal_duration = args.enforce_signal_duration

        self.run_local = args.local
        self.local_generation = args.local_generation
        self.local_plot = args.local_plot

        self.post_trigger_duration = args.post_trigger_duration

        # self.ignore_gwpy_data_quality_check = args.ignore_gwpy_data_quality_check
        self.trigger_time = args.trigger_time
        # self.deltaT = args.deltaT
        # self.gps_tuple = args.gps_tuple
        # self.gps_file = args.gps_file
        self.timeslide_file = args.timeslide_file
        # self.gaussian_noise = args.gaussian_noise
        # self.zero_noise = args.zero_noise
        # self.n_simulation = args.n_simulation
        #
        # self.injection = args.injection
        # self.injection_numbers = args.injection_numbers
        self.injection_file = args.injection_file
        # self.injection_dict = args.injection_dict
        # self.injection_waveform_arguments = args.injection_waveform_arguments
        # self.injection_waveform_approximant = args.injection_waveform_approximant
        # self.generation_seed = args.generation_seed
        # if self.injection:
        #     self.check_injection()

        self.importance_sample = args.importance_sample

        self.request_disk = args.request_disk
        self.request_memory = args.request_memory
        self.request_memory_generation = args.request_memory_generation
        self.request_cpus = args.request_cpus
        self.request_cpus_importance_sampling = args.request_cpus_importance_sampling
        # self.sampler_kwargs = args.sampler_kwargs
        # self.mpi_samplers = ["pymultinest"]
        # self.use_mpi = (self.sampler in self.mpi_samplers) and (self.request_cpus > 1)
        #
        # # Set plotting options when need the plot node
        self.plot_node_needed = False
        for plot_attr in [
            "corner",
            "weights",
            "log_probs",
        ]:
            attr = f"plot_{plot_attr}"
            setattr(self, attr, getattr(args, attr))
            if getattr(self, attr):
                self.plot_node_needed = True
        #
        # # Set all other plotting options
        # for plot_attr in [
        #     "trace",
        #     "data",
        #     "injection",
        #     "spectrogram",
        #     "format",
        # ]:
        #     attr = f"plot_{plot_attr}"
        #     setattr(self, attr, getattr(args, attr))
        #
        # self.postprocessing_executable = args.postprocessing_executable
        # self.postprocessing_arguments = args.postprocessing_arguments
        # self.single_postprocessing_executable = args.single_postprocessing_executable
        # self.single_postprocessing_arguments = args.single_postprocessing_arguments
        #
        self.summarypages_arguments = args.summarypages_arguments
        #
        self.psd_dict = args.psd_dict

        # self.check_source_model(args)

        self.requirements = []
        self.device = args.device
        self.simple_submission = args.simple_submission

        if args.extra_lines:
            self.extra_lines = args.extra_lines
        else:
            self.extra_lines = []

        if args.sampling_requirements:
            self.sampling_requirements = args.sampling_requirements
        else:
            self.sampling_requirements = []

    @property
    def request_cpus_importance_sampling(self):
        return self._request_cpus_importance_sampling

    @request_cpus_importance_sampling.setter
    def request_cpus_importance_sampling(self, request_cpus_importance_sampling):
        logger.info(
            f"Setting analysis request_cpus_importance_sampling = "
            f"{request_cpus_importance_sampling}"
        )
        self._request_cpus_importance_sampling = request_cpus_importance_sampling

    @property
    def priors(self):
        """Read in and compose the prior at run-time"""
        if getattr(self, "_priors", None) is None:
            self._priors = self._get_priors(add_time=False)
        return self._priors


def write_complete_config_file(parser, args, inputs, input_cls=MainInput):
    args_dict = vars(args).copy()
    for key, val in args_dict.items():
        if key == "label":
            continue
        if isinstance(val, str):
            if os.path.isfile(val) or os.path.isdir(val):
                setattr(args, key, os.path.abspath(val))
        if isinstance(val, list):
            if isinstance(val[0], str):
                setattr(args, key, f"[{', '.join(val)}]")
    # args.sampler_kwargs = str(inputs.sampler_kwargs)
    args.importance_sampling_updates = str(inputs.importance_sampling_updates)
    args.submit = False
    parser.write_to_file(
        filename=inputs.complete_ini_file,
        args=args,
        overwrite=False,
        include_description=False,
    )

    # Verify that the written complete config is identical to the source config
    # complete_args = parser.parse([inputs.complete_ini_file])
    # complete_inputs = input_cls(complete_args, "")
    # ignore_keys = ["scheduler_module", "submit"]
    # differences = []
    # for key, val in inputs.__dict__.items():
    #     if key in ignore_keys:
    #         continue
    #     if key not in complete_args:
    #         continue
    #     if isinstance(val, pd.DataFrame) and all(val == complete_inputs.__dict__[key]):
    #         continue
    #     if isinstance(val, np.ndarray) and all(
    #         np.array(val) == np.array(complete_inputs.__dict__[key])
    #     ):
    #         continue
    #     if isinstance(val, str) and os.path.isfile(val):
    #         # Check if it is relpath vs abspath
    #         if os.path.abspath(val) == os.path.abspath(complete_inputs.__dict__[key]):
    #             continue
    #     if val == complete_inputs.__dict__[key]:
    #         continue
    #
    #     differences.append(key)
    #
    # if len(differences) > 0:
    #     for key in differences:
    #         print(key, f"{inputs.__dict__[key]} -> {complete_inputs.__dict__[key]}")
    # raise BilbyPipeError(
    #     "The written config file {} differs from the source {} in {}".format(
    #         inputs.ini, inputs.complete_ini_file, differences
    #     )
    # )


def main():
    parser = create_parser(top_level=True)
    args, unknown_args = parse_args(get_command_line_arguments(), parser)

    importance_sampling_updates, model_args = fill_in_arguments_from_model(args)
    inputs = MainInput(args, unknown_args, importance_sampling_updates)
    write_complete_config_file(parser, args, inputs)

    # TODO: Use two sets of inputs! The first must match the network; the second is
    #  used in importance sampling.

    generate_dag(inputs, model_args)

    if len(unknown_args) > 0:
        print(f"Unrecognized arguments {unknown_args}")



import argparse
import os
import pickle
from os.path import join

import numpy as np
import yaml
from tqdm import tqdm

from dingo.gw.domains import build_domain
from dingo.gw.download_strain_data import estimate_single_psd
from dingo.gw.gwutils import get_window
from dingo.gw.noise.asd_dataset import ASDDataset
from dingo.gw.noise.utils import psd_data_path


def download_and_estimate_psds(
    data_dir: str,
    settings: dict,
    time_segments: dict,
    verbose=False,
):
    """
    Downloads strain data for the specified time segments and estimates PSDs based on these

    Parameters
    ----------
    data_dir : str
        Path to the directory where the PSD dataset will be stored
    settings : dict
        Settings that determine the segments
    time_segments : dict
        specifying the time segments used for downloading the data
    verbose : bool
        optional parameter determining if progress should be printed
    Returns
    -------
    A dictionary containing the paths to the dataset files
    """
    dataset_settings = settings["dataset_settings"]
    run = dataset_settings["observing_run"]
    detectors = (
        time_segments.keys()
    )  # time_segments may only contain a subset of all detectors for parallelization

    f_min = dataset_settings.get("f_min", 0)
    f_max = dataset_settings.get("f_max", (dataset_settings["f_s"] / 2))
    T = dataset_settings["T"]
    time_psd = dataset_settings["time_psd"]
    f_s = dataset_settings["f_s"]

    channels = dataset_settings.get("channels", None)

    delta_f = 1 / T
    domain = build_domain(
        {
            "type": "FrequencyDomain",
            "f_min": f_min,
            "f_max": f_max,
            "delta_f": delta_f,
            "window_factor": None,
        }
    )

    window_kwargs = {
        "f_s": f_s,
        "roll_off": dataset_settings["window"]["roll_off"],
        "type": dataset_settings["window"]["type"],
        "T": T,
    }
    w = get_window(window_kwargs)
    asd_filename_list = {det: [] for det in detectors}
    for det in detectors:
        psd_path = psd_data_path(data_dir, run, det)
        os.makedirs(psd_path, exist_ok=True)
        estimation_kwargs = {
            "time_segment": T,
            "window": w,
            "f_s": f_s,
            "num_segments": int(time_psd / T),
        }
        if channels:
            estimation_kwargs["channel"] = channels[det]
        else:
            estimation_kwargs["det"] = det

        for index, (start, end) in enumerate(
            tqdm(time_segments[det], disable=not verbose)
        ):
            filename = join(psd_path, f"asd_{start}.hdf5")
            asd_filename_list[det].append(filename)
            if not os.path.exists(filename):
                dataset_dict = {
                    "settings": {
                        "dataset_settings": settings["dataset_settings"],
                        "domain_dict": domain.domain_dict,
                    }
                }
                psd = estimate_single_psd(time_start=start, **estimation_kwargs)
                asd = np.sqrt(psd[domain.min_idx : domain.max_idx + 1])
                gps_time = start

                dataset_dict["asds"] = {det: np.array([asd])}
                dataset_dict["gps_times"] = {det: np.array([gps_time])}

                dataset = ASDDataset(dictionary=dataset_dict)
                dataset.to_file(file_name=filename)

    return asd_filename_list


def download_and_estimate_cli():
    """
    Command-line function to download strain data and estimate PSDs based on the data. Used for
    parallelized ASD dataset generation.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--data_dir",
        type=str,
        required=True,
        help="Path where the PSD data is to be stored.",
    )
    parser.add_argument(
        "--settings_file",
        type=str,
        required=True,
        help="Path to a settings file containing the settings for the dataset generation",
    )
    parser.add_argument(
        "--time_segments_file",
        type=str,
        default=None,
        help="Path to a file containing the time segments for which PSDs should be estimated",
    )

    args = parser.parse_args()

    # Load settings
    with open(args.settings_file, "r") as f:
        settings = yaml.safe_load(f)

    with open(args.time_segments_file, "rb") as f:
        time_segments = pickle.load(f)

    download_and_estimate_psds(args.data_dir, settings, time_segments)



import os
import uuid

import numpy as np
import pytest

from dingo.gw.domains import Domain
from dingo.gw.dataset.waveform_dataset import WaveformDataset
from dingo.gw.dataset.generate_dataset_dag import create_args_string

SETTINGS_YAML_SMALL = """\
# settings for domain of waveforms
domain:
  type: FrequencyDomain
  f_min: 10.0
  f_max: 1024.0
  delta_f: 1.0

# settings for waveform generator
waveform_generator:
  approximant: IMRPhenomPv2
  f_ref: 20.0

# settings for intrinsic prior over parameters
intrinsic_prior:
  # prior for non-fixed parameters
  mass_1: bilby.core.prior.Constraint(minimum=10.0, maximum=80.0)
  mass_2: bilby.core.prior.Constraint(minimum=10.0, maximum=80.0)
  mass_ratio: bilby.core.prior.Uniform(minimum=0.125, maximum=1.0)
  chirp_mass: bilby.core.prior.Uniform(minimum=25.0, maximum=100.0)
  phase: default
  a_1: bilby.core.prior.Uniform(minimum=0.0, maximum=0.88)
  a_2: bilby.core.prior.Uniform(minimum=0.0, maximum=0.88)
  tilt_1: default
  tilt_2: default
  phi_12: default
  phi_jl: default
  theta_jn: default
  # reference values for fixed (extrinsic) parameters
  luminosity_distance: 100.0 # Mpc
  geocent_time: 0.0 # s

num_samples: 50

compression:
  svd:
    num_training_samples: 10
    size: 5
"""


@pytest.fixture
def generate_waveform_dataset_small(venv_dir='venv'):
    """
    Create a small waveform dataset on-the-fly
    in a temporary directory.

    venv_dir:
        The name of the folder that contains the dingo
        virtual environment relative to the 'dingo-devel' root.
    """
    # Figure out the path to the 'dingo-devel' root directory
    # s = os.getcwd()
    # root = s.split('dingo-devel')[0]  # + 'dingo-devel'
    # venv_path = os.path.join(root, venv_dir)
    # generate_waveform_path = os.path.join(venv_path, 'bin/dingo_generate_waveforms')

    # Create temp directory and settings file
    path = os.path.join('./tmp_test', str(uuid.uuid4()))
    os.makedirs(path, exist_ok=True)
    with open(os.path.join(path, 'settings.yaml'), 'w') as fp:
        fp.writelines(SETTINGS_YAML_SMALL)

    # Mock up command line arguments for script
    args_in = {'settings_file': os.path.join(path, 'settings.yaml'),
               'out_file': os.path.join(path, 'waveform_dataset.hdf5'),
               'num_processes': '4'}
    args_string = create_args_string(args_in)
    res = os.system('dingo_generate_dataset ' + args_string)
    if not os.WIFEXITED(res):
        raise RuntimeError(f'dingo_generate_waveforms returned a '
                           f'nonzero exit code: {res}')

    # args = parse_args(args_in)
    # out_script = './waveform_generation_script.sh'
    # if os.path.exists(out_script):
    #     os.remove(out_script)
    # generate_workflow(args)
    #
    # # Now execute the generated workflow
    # os.chmod(out_script, stat.S_IREAD | stat.S_IWRITE | stat.S_IXUSR)
    # res = os.system(f'{out_script} > ./waveform_generation_script.log 2>&1')
    # if not os.WIFEXITED(res):
    #     raise RuntimeError(f'waveform_generation_script.sh returned a '
    #                        f'nonzero exit code: {res}')

    return path

@pytest.mark.slow
def test_load_waveform_dataset(generate_waveform_dataset_small):
    wfd_path = generate_waveform_dataset_small

    path = f'{wfd_path}/waveform_dataset.hdf5'
    wd = WaveformDataset(file_name=path, precision='single')

    assert len(wd) > 0
    el = wd[0]
    assert isinstance(el, dict)
    assert set(el.keys()) == {'parameters', 'waveform'}
    assert isinstance(el['parameters'], dict)
    assert isinstance(el['waveform'], dict)
    assert isinstance(el['waveform']['h_plus'], np.ndarray)
    assert isinstance(el['waveform']['h_cross'], np.ndarray)

    # Check the associated domain
    assert len(wd.domain) > 0
    assert isinstance(wd.domain, Domain)
    assert isinstance(wd.domain.domain_dict, dict)

    # Check the associated data settings
    data_settings_keys = {'domain', 'waveform_generator',
                          'intrinsic_prior', 'num_samples',
                          'compression'}
    assert set(wd.settings.keys()) == data_settings_keys


    """Check truncation of wd. Ideally, this should be an individual test, 
    but since the setup takes so long it is added here."""

    f_min = wd.domain.f_min
    f_max = wd.domain.f_max
    delta_f = wd.domain._delta_f

    # check that truncation works as intended when setting new range
    f_min_new = 20
    f_max_new = 100
    wd2 = WaveformDataset(path, domain_update={'f_min': f_min_new, 'f_max': f_max_new})
    assert len(wd2.domain) == len(wd2.domain())
    # check that new domain settings are correctly adapted
    assert wd2.domain.f_min == f_min_new
    assert wd2.domain.f_max == f_max_new
    assert wd2.domain._delta_f == wd.domain._delta_f
    # check that truncation works as intended
    for pol in ['h_cross', 'h_plus']:
        # f_min_new to f_max_new check
        a = el['waveform'][pol][int(f_min_new/delta_f):int(f_max_new/delta_f)+1]
        b = wd2[0]['waveform'][pol][int(f_min_new/delta_f):]
        scale_factor = np.max(np.abs(a))
        assert len(a) == f_max_new / delta_f + 1 - f_min_new / delta_f
        assert np.allclose(b / scale_factor, a / scale_factor)
        assert not np.allclose(b / scale_factor, np.roll(a, 1) / scale_factor)

        # f_min to f_min_new check
        a = el['waveform'][pol][int(f_min / delta_f):int(f_min_new / delta_f)]
        b = wd2[0]['waveform'][pol][int(f_min / delta_f):int(f_min_new / delta_f)]
        assert not np.allclose(b / scale_factor, a / scale_factor)

        # below f_min_new check
        assert np.all(wd2[0]['waveform'][pol][:int(f_min_new)] == 0.0)
    assert len(wd2.domain) == f_max_new / delta_f + 1
    assert len(wd2.domain) == len(wd2.domain())


import os

import torch
import argparse
import yaml
import h5py
import ast


def parse_args():

    parser = argparse.ArgumentParser()
    parser.add_argument('--checkpoint', type=str, required=True)
    parser.add_argument('--settings_file', type=str, required=True)
    parser.add_argument('--out_file', type=str, required=True)
    parser.add_argument('--train_dir', type=str, required=True)
    return parser.parse_args()


def main():
    args = parse_args()

    with open(args.settings_file, 'r') as f:
        train_settings = yaml.safe_load(f)

    # Typically training is done on the GPU, so the model could be saved on a GPU
    # device. Since this routine may be run on a CPU machine, allow for a remap of the
    # torch tensors.
    if torch.cuda.is_available():
        d = torch.load(args.checkpoint)
    else:
        d = torch.load(args.checkpoint, map_location=torch.device('cpu'))

    data = {
        'waveform_dataset_path': train_settings['waveform_dataset_path'],
        'train_fraction': train_settings['train_settings']['train_fraction'],
        'conditioning': train_settings['data_conditioning'],
    }
    data.update(train_settings['transform_settings'])

    model = d['model_kwargs']
    model['type'] = 'nsf+embedding'
    model['embedding_net_kwargs']['svd'] = {
        'size': model['embedding_net_kwargs']['n_rb'],
    },
    del model['embedding_net_kwargs']['n_rb']
    del model['embedding_net_kwargs']['V_rb_list']

    training = {'stage_0': {
        'epochs': train_settings['train_settings']['runtime_limits']['max_epochs_total'],
        'asd_dataset_path': train_settings['asd_dataset_path'],
        'freeze_rb_layer': train_settings['train_settings']['freeze_rb_layer'],
        'optimizer': train_settings['train_settings']['optimizer_kwargs'],
        'scheduler': train_settings['train_settings']['scheduler_kwargs'],
        'batch_size': train_settings['train_settings']['batch_size'],
    }
    }

    local = {
        'device': train_settings['train_settings']['device'],
        'num_workers': train_settings['train_settings']['num_workers'],
        'runtime_limits': train_settings['train_settings']['runtime_limits'],
        'checkpoint_epochs': train_settings['train_settings']['checkpoint_epochs'],
        'condor': train_settings['condor_settings'],
    }
    del local['runtime_limits']['max_epochs_total']

    d['metadata'] = {'train_settings': {
        'data': data,
        'model': model,
        'training': training,
    }}

    # Save (for posterity) the waveform dataset settings
    f = h5py.File(d['metadata']['train_settings']['data']['waveform_dataset_path'], 'r')
    settings = ast.literal_eval(f.attrs['settings'])
    d['metadata']['dataset_settings'] = settings
    f.close()

    torch.save(d, args.out_file)

    # Save local settings
    with open(os.path.join(args.train_dir, 'local_settings.yaml'), 'w') as f:
        yaml.dump(local, f, default_flow_style=False, sort_keys=False)


if __name__ == "__main__":
    main()



import copy
import time
from typing import Optional

import numpy as np
import yaml
from bilby.core.prior import Uniform, Constraint, PriorDict

from dingo.core.density import (
    interpolated_sample_and_log_prob_multi,
    interpolated_log_prob_multi,
)
from dingo.core.multiprocessing import apply_func_with_multiprocessing
from dingo.core.result import Result as CoreResult
from dingo.gw.conversion import change_spin_conversion_phase
from dingo.gw.domains import build_domain
from dingo.gw.gwutils import get_extrinsic_prior_dict, get_window_factor
from dingo.gw.likelihood import StationaryGaussianGWLikelihood
from dingo.gw.prior import build_prior_with_defaults


RANDOM_STATE = 150914


class Result(CoreResult):
    """
    A dataset class to hold a collection of gravitational-wave parameter samples and
    perform various operations on them.

    Compared to the base class, this class implements the domain, prior,
    and likelihood. It also includes a method for sampling the binary reference phase
    parameter based on the other parameters and the likelihood.

    Attributes:
        samples : pd.Dataframe
            Contains parameter samples, as well as (possibly) log_prob, log_likelihood,
            weights, log_prior, delta_log_prob_target.
        domain : Domain
            The domain of the data (e.g., FrequencyDomain), needed for calculating
            likelihoods.
        prior : PriorDict
            The prior distribution, used for importance sampling.
        likelihood : Likelihood
            The Likelihood object, needed for importance sampling.
        context : dict
            Context data from which the samples were produced (e.g., strain data, ASDs).
        metadata : dict
            Metadata inherited from the Sampler object. This describes the neural
            networks and sampling settings used.
        event_metadata : dict
            Metadata for the event analyzed, including time, data conditioning, channel,
            and detector information.
        log_evidence : float
            Calculated log(evidence) after importance sampling.
        log_evidence_std : float (property)
            Standard deviation of the log(evidence)
        effective_sample_size, n_eff : float (property)
            Number of effective samples, (\\sum_i w_i)^2 / \\sum_i w_i^2
        sample_efficiency : float (property)
            Number of effective samples / Number of samples
        synthetic_phase_kwargs : dict
            kwargs describing the synthetic phase sampling.
    """

    dataset_type = "gw_result"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def synthetic_phase_kwargs(self):
        return self.importance_sampling_metadata.get("synthetic_phase")

    @synthetic_phase_kwargs.setter
    def synthetic_phase_kwargs(self, value):
        self.importance_sampling_metadata["synthetic_phase"] = value

    @property
    def time_marginalization_kwargs(self):
        return self.importance_sampling_metadata.get("time_marginalization")

    @time_marginalization_kwargs.setter
    def time_marginalization_kwargs(self, value):
        self.importance_sampling_metadata["time_marginalization"] = value

    @property
    def phase_marginalization_kwargs(self):
        return self.importance_sampling_metadata.get("phase_marginalization")

    @phase_marginalization_kwargs.setter
    def phase_marginalization_kwargs(self, value):
        self.importance_sampling_metadata["phase_marginalization"] = value

    @property
    def calibration_marginalization_kwargs(self):
        return self.importance_sampling_metadata.get("calibration_marginalization")

    @calibration_marginalization_kwargs.setter
    def calibration_marginalization_kwargs(self, value):
        self.importance_sampling_metadata["calibration_marginalization"] = value

    @property
    def f_ref(self):
        return self.base_metadata["dataset_settings"]["waveform_generator"]["f_ref"]

    @property
    def approximant(self):
        return self.base_metadata["dataset_settings"]["waveform_generator"][
            "approximant"
        ]

    @property
    def interferometers(self):
        return list(self.context["waveform"].keys())

    @property
    def t_ref(self):
        # The detector reference positions during likelihood evaluation should be
        # based on the event time, since any post-correction to account for the training
        # reference time has already been applied to the samples.
        if self.event_metadata is not None and "time_event" in self.event_metadata:
            return self.event_metadata["time_event"]
        else:
            return self.base_metadata["train_settings"]["data"]["ref_time"]

    def _build_domain(self):
        """
        Construct the domain object based on model metadata. Includes the window factor
        needed for whitening data.

        Called by __init__() immediately after _build_prior().
        """
        self.domain = build_domain(self.base_metadata["dataset_settings"]["domain"])

        data_settings = self.base_metadata["train_settings"]["data"]
        if "domain_update" in data_settings:
            self.domain.update(data_settings["domain_update"])

        self.domain.window_factor = get_window_factor(data_settings["window"])

    def _rebuild_domain(self, verbose=False):
        """Rebuild the domain based on settings updated for importance sampling.

        These settings should be saved in self.importance_sampling_metadata["updates"],
        which is expected to be populated by reset_event()."""
        updates = self.importance_sampling_metadata["updates"].copy()

        # Assume that updates can contain T, f_s, roll_off, f_min, f_max, but no other
        # quantities that define a new domain (e.g., delta_f). Typical event metadata
        # will be constructed in this way.

        if "f_s" in updates or "T" in updates or "roll_off" in updates:
            window_settings = self.base_metadata["train_settings"]["data"][
                "window"
            ].copy()
            window_settings.update(
                (k, updates[k]) for k in set(window_settings).intersection(updates)
            )
            updates["window_factor"] = float(get_window_factor(window_settings))

        if "T" in updates:
            updates["delta_f"] = 1.0 / updates["T"]

        domain_dict = self.domain.domain_dict  # Existing settings
        domain_dict.update(
            (k, updates[k]) for k in set(domain_dict).intersection(updates)
        )

        if verbose:
            print("Rebuilding domain as follows:")
            print(
                yaml.dump(
                    domain_dict,
                    default_flow_style=False,
                    sort_keys=False,
                )
            )
        self.domain = build_domain(domain_dict)

    def _build_prior(self):
        """Build the prior based on model metadata. Called by __init__()."""
        intrinsic_prior = self.base_metadata["dataset_settings"]["intrinsic_prior"]
        extrinsic_prior = get_extrinsic_prior_dict(
            self.base_metadata["train_settings"]["data"]["extrinsic_prior"]
        )
        self.prior = build_prior_with_defaults({**intrinsic_prior, **extrinsic_prior})

        prior_update = self.importance_sampling_metadata.get("prior_update")
        if prior_update is not None:
            prior_update = PriorDict(prior_update.copy())
            self.prior.update(prior_update)

        # Split off prior over geocent_time if samples appear to be time-marginalized.
        # This needs to be saved to initialize the likelihood.
        if "geocent_time" in self.prior.keys() and "geocent_time" not in self.samples:
            self.geocent_time_prior = self.prior.pop("geocent_time")
        else:
            self.geocent_time_prior = None
        # Split off prior over phase if samples appear to be phase-marginalized.
        if "phase" in self.prior.keys() and "phase" not in self.samples:
            self.phase_prior = self.prior.pop("phase")
        else:
            self.phase_prior = None

    def update_prior(self, prior_update):
        """
        Update the prior based on a new dict of priors. Use the existing prior for
        parameters not included in the new dict.

        If class samples have not been importance sampled, then save new sample weights
        based on the new prior. If class samples have been importance sampled,
        then update the weights.

        Parameters
        ----------
        prior_update : dict
            Priors to update. This should be of the form {key : prior_str}, where str
            is a string that can instantiate a prior via PriorDict(prior_update). The
            prior_update is provided in this form so that it can be properly saved with
            the Result and later instantiated.
        """
        self.importance_sampling_metadata["prior_update"] = prior_update.copy()
        prior_update = PriorDict(prior_update)

        param_keys = [k for k, v in self.prior.items() if not isinstance(v, Constraint)]
        theta = self.samples[param_keys]

        if self.log_evidence is None:
            # Save old prior evaluations.
            log_prior_old = self.prior.ln_prob(theta, axis=0)

        # Update the prior itself, careful to split off geocent_time and phase priors
        # if necessary.
        if self.geocent_time_prior is not None and "geocent_time" in prior_update:
            self.geocent_time_prior = prior_update.pop("geocent_time")
        if self.phase_prior is not None and "phase" in prior_update:
            self.phase_prior = prior_update.pop("phase")
        self.prior.update(
            prior_update
        )  # TODO: Does this update cached constraint ratio?

        # Evaluate new prior.
        log_prior = self.prior.ln_prob(theta, axis=0)
        self.samples["log_prior"] = log_prior

        if self.log_evidence is None:
            # Save weights. Note that weights are 0 if outside the initial prior,
            # regardless of new prior. This makes sense since there is no way to assign
            # a sensible weight.
            log_weights = log_prior - log_prior_old
            weights = np.exp(log_weights - np.max(log_weights))
            weights /= np.mean(weights)
            self.samples["weights"] = weights

        else:
            # Recalculate the importance-sampling weights and log evidence.
            self._calculate_evidence()

    def _build_likelihood(
        self,
        time_marginalization_kwargs: Optional[dict] = None,
        phase_marginalization_kwargs: Optional[dict] = None,
        calibration_marginalization_kwargs: Optional[dict] = None,
        phase_grid: Optional[np.ndarray] = None,
    ):
        """
        Build the likelihood function based on model metadata. This is called at the
        beginning of importance_sample().

        Parameters
        ----------
        time_marginalization_kwargs: dict, optional
            kwargs for time marginalization. At this point the only kwarg is n_fft,
            which determines the number of FFTs used (higher n_fft means better
            accuracy, at the cost of longer computation time).
        phase_marginalization_kwargs: dict, optional
            kwargs for phase marginalization.
        calibration_marginalization_kwargs: dict
            Calibration marginalization parameters. If None, no calibration marginalization is used.
        """
        if time_marginalization_kwargs is not None:
            if self.geocent_time_prior is None:
                raise NotImplementedError(
                    "Time marginalization is not compatible with "
                    "non-marginalized network."
                )
            if type(self.geocent_time_prior) != Uniform:
                raise NotImplementedError(
                    "Only uniform time prior is supported for time marginalization."
                )
            time_marginalization_kwargs["t_lower"] = self.geocent_time_prior.minimum
            time_marginalization_kwargs["t_upper"] = self.geocent_time_prior.maximum

        if phase_marginalization_kwargs is not None:
            # check that phase prior is uniform [0, 2pi)
            if not (
                isinstance(self.phase_prior, Uniform)
                and (self.phase_prior._minimum, self.phase_prior._maximum)
                == (0, 2 * np.pi)
            ):
                raise ValueError(
                    f"Phase prior should be uniform [0, 2pi) for phase "
                    f"marginalization, but is {self.phase_prior}."
                )

        # This will save these settings when the Result instance is saved.
        self.time_marginalization_kwargs = time_marginalization_kwargs
        self.phase_marginalization_kwargs = phase_marginalization_kwargs
        self.calibration_marginalization_kwargs = calibration_marginalization_kwargs

        # FIXME: This is a quick hack because I didn't know how to choose the wfg
        #  domain in the case of a changing domain during importance sampling. It could
        #  only pose problems for EOB, where sometimes one wants to start integrating
        #  from lower f_min. But even in that case, f_start should override it.
        if self.importance_sampling_metadata.get("updates") is None:
            wfg_domain = build_domain(self.base_metadata["dataset_settings"]["domain"])
        else:
            wfg_domain = self.domain

        self.likelihood = StationaryGaussianGWLikelihood(
            wfg_kwargs=self.base_metadata["dataset_settings"]["waveform_generator"],
            wfg_domain=wfg_domain,
            data_domain=self.domain,
            event_data=self.context,
            t_ref=self.t_ref,
            time_marginalization_kwargs=time_marginalization_kwargs,
            phase_marginalization_kwargs=phase_marginalization_kwargs,
            calibration_marginalization_kwargs=calibration_marginalization_kwargs,
            phase_grid=phase_grid,
        )

    def sample_synthetic_phase(
        self,
        synthetic_phase_kwargs,
        inverse: bool = False,
    ):
        """
        Sample a synthetic phase for the waveform. This is a post-processing step
        applicable to samples theta in the full parameter space, except for the phase
        parameter (i.e., 14D samples). This step adds a phase parameter to the samples
        based on likelihood evaluations.

        A synthetic phase is sampled as follows.

            * Compute and cache the modes for the waveform mu(theta, phase=0) for
              phase 0, organize them such that each contribution m transforms as
              exp(-i * m * phase).
            * Compute the likelihood on a phase grid, by computing mu(theta, phase) from
              the cached modes. In principle this likelihood is exact, however, it can
              deviate slightly from the likelihood computed without cached modes for
              various technical reasons (e.g., slightly different windowing of cached
              modes compared to full waveform when transforming TD waveform to FD).
              These small deviations can be fully accounted for by importance sampling.
              *Note*: when approximation_22_mode=True, the entire waveform is assumed
              to transform as exp(2i*phase), in which case the likelihood is only exact
              if the waveform is fully dominated by the (2, 2) mode.
            * Build a synthetic conditional phase distribution based on this grid. We
              use an interpolated prior distribution bilby.core.prior.Interped,
              such that we can sample and also evaluate the log_prob. We add a constant
              background with weight self.synthetic_phase_kwargs to the kde to make
              sure that we keep a mass-covering property. With this, the importance
              sampling will yield exact results even when the synthetic phase conditional
              is just an approximation.

        Besides adding phase samples to self.samples['phase'], this method also modifies
        self.samples['log_prob'] by adding the log_prob of the synthetic phase
        conditional.

        This method modifies self.samples in place.

        Parameters
        ----------
        synthetic_phase_kwargs : dict
            This should consist of the kwargs
                approximation_22_mode (optional)
                num_processes (optional)
                n_grid
                uniform_weight (optional)
        inverse : bool, default False
            Whether to apply instead the inverse transformation. This is used prior to
            calculating the log_prob. In inverse mode, the posterior probability over
            phase is calculated for given samples. It is stored in self.samples[
            'log_prob'].
        """
        self.synthetic_phase_kwargs = synthetic_phase_kwargs

        approximation_22_mode = self.synthetic_phase_kwargs.get(
            "approximation_22_mode", True
        )

        if not (
            isinstance(self.phase_prior, Uniform)
            and (self.phase_prior._minimum, self.phase_prior._maximum) == (0, 2 * np.pi)
        ):
            raise ValueError(
                f"Phase prior should be uniform [0, 2pi) to work with synthetic phase."
                f" However, the prior is {self.phase_prior}."
            )

        # This builds on the Bilby approach to sampling the phase when using a
        # phase-marginalized likelihood.

        # Restrict to samples that are within the prior.
        param_keys = [k for k, v in self.prior.items() if not isinstance(v, Constraint)]
        theta = self.samples[param_keys]
        log_prior = self.prior.ln_prob(theta, axis=0)
        constraints = self.prior.evaluate_constraints(theta)
        np.putmask(log_prior, constraints == 0, -np.inf)
        within_prior = log_prior != -np.inf

        # Put a cap on the number of processes to avoid overhead:
        num_valid_samples = np.sum(within_prior)
        num_processes = min(
            self.synthetic_phase_kwargs.get("num_processes", 1), num_valid_samples // 10
        )

        print(f"Estimating synthetic phase for {num_valid_samples} samples.")
        t0 = time.time()

        if not inverse:
            # TODO: This can probably be removed.
            self._build_likelihood()

        if inverse:
            # We estimate the log_prob for given phases, so first save the evaluation
            # points.
            sample_phase = theta["phase"].to_numpy(copy=True)

        # For each sample, build the posterior over phase given the remaining parameters.

        phases = np.linspace(0, 2 * np.pi, self.synthetic_phase_kwargs["n_grid"])
        if approximation_22_mode:
            # For each sample, the un-normalized posterior depends only on (d | h(phase)):
            # The prior p(phase), and the inner products (h | h), and (d | d) only contribute
            # to the normalization. (We check above that p(phase) is constant.)
            theta["phase"] = 0.0
            d_inner_h_complex = self.likelihood.d_inner_h_complex_multi(
                theta.iloc[within_prior],
                num_processes,
            )

            # Evaluate the log posterior over the phase across the grid.
            phasor = np.exp(2j * phases)
            phase_log_posterior = np.outer(d_inner_h_complex, phasor).real
        else:
            self.likelihood.phase_grid = phases

            phase_log_posterior = apply_func_with_multiprocessing(
                self.likelihood.log_likelihood_phase_grid,
                theta.iloc[within_prior],
                num_processes=num_processes,
            )

        phase_posterior = np.exp(
            phase_log_posterior - np.amax(phase_log_posterior, axis=1, keepdims=True)
        )
        # Include a floor value to maintain mass coverage.
        phase_posterior += phase_posterior.mean(
            axis=-1, keepdims=True
        ) * self.synthetic_phase_kwargs.get("uniform_weight", 0.01)

        if not inverse:
            # Forward direction:
            #   (1) Sample a new phase according to the synthetic posterior.
            #   (2) Add the log_prob to the existing log_prob.

            new_phase, delta_log_prob = interpolated_sample_and_log_prob_multi(
                phases,
                phase_posterior,
                num_processes,
            )

            phase_array = np.full(len(theta), 0.0)
            phase_array[within_prior] = new_phase
            delta_log_prob_array = np.full(len(theta), -np.nan)
            delta_log_prob_array[within_prior] = delta_log_prob

            self.samples["phase"] = phase_array
            self.samples["log_prob"] += delta_log_prob_array

            # Insert the phase prior in the prior, since now the phase is present.
            self.prior["phase"] = self.phase_prior
            self.phase_prior = None

            # reset likelihood for safety
            # TODO: Can this be removed?
            self.likelihood = None

        else:
            # TODO: Possibly remove.
            # Inverse direction:
            #   (1) Evaluate the synthetic log prob for given phase points, and save it.

            log_prob = interpolated_log_prob_multi(
                phases,
                phase_posterior,
                sample_phase[within_prior],
                num_processes,
            )

            # Outside of prior, set log_prob to -np.nan.
            log_prob_array = np.full(len(theta), -np.nan)
            log_prob_array[within_prior] = log_prob
            self.samples["log_prob"] = log_prob_array
            del self.samples["phase"]

        print(f"Done. This took {time.time() - t0:.2f} s.")

    def get_samples_bilby_phase(self):
        """
        Convert the spin angles phi_jl and theta_jn to account for a difference in
        phase definition compared to Bilby.

        Returns
        -------
        pd.DataFrame
            Samples
        """
        spin_conversion_phase_old = self.base_metadata["dataset_settings"][
            "waveform_generator"
        ].get("spin_conversion_phase")

        # Redefine phase parameter to be consistent with Bilby.
        return change_spin_conversion_phase(
            self.samples, self.f_ref, spin_conversion_phase_old, None
        )

    @property
    def pesummary_samples(self):
        """Samples in a form suitable for PESummary.

        These samples are adjusted to undo certain conventions used internally by
        Dingo:
            * Times are corrected by the reference time t_ref.
            * Samples are unweighted, using a fixed random seed for sampling importance
            resampling.
            * The spin angles phi_jl and theta_jn are transformed to account for a
            difference in phase definition.
            * Some columns are dropped: delta_log_prob_target, log_prob
        """
        # Unweighted samples.
        samples = self.sampling_importance_resampling(random_state=RANDOM_STATE)

        # Remove unwanted columns.
        samples.drop(
            ["delta_log_prob_target", "log_prob"], axis=1, errors="ignore", inplace=True
        )
        for col in samples.columns:
            if col.endswith("_proxy"):
                samples.drop(col, axis=1, inplace=True)

        # Shift times. This requires double precision. There *should* be no non-numeric
        # values in the samples dataframe since resampling will have excluded
        # zero-weight samples (which could have nan likelihood).
        samples = samples.astype(float)
        for col in samples.columns:
            if "time" in col:
                samples.loc[:, col] += self.t_ref

        spin_conversion_phase_old = self.base_metadata["dataset_settings"][
            "waveform_generator"
        ].get("spin_conversion_phase")

        # Redefine phase parameter to be consistent with Bilby. COMMENTED BECAUSE SLOW
        samples = change_spin_conversion_phase(
            samples, self.f_ref, spin_conversion_phase_old, None
        )

        return samples

    @property
    def pesummary_prior(self):
        """The prior in a form suitable for PESummary.

        By convention, Dingo stores all times *relative* to a reference time, typically
        the trigger time for an event. The prior returned here corrects for that offset to
        be consistent with other codes.
        """
        prior = copy.deepcopy(self.prior)
        for p in prior:
            if "time" in p:
                try:
                    prior[p].maximum += self.t_ref
                    prior[p].minimum += self.t_ref
                except AttributeError:
                    continue
        return prior



#
#  Adapted from bilby_pipe.
#

import copy

from bilby_pipe.job_creation.dag import Dag
from bilby_pipe.utils import BilbyPipeError, logger

from dingo.pipe.nodes.generation_node import GenerationNode
from .nodes.importance_sampling_node import ImportanceSamplingNode
from .nodes.merge_node import MergeNode
from .nodes.pe_summary_node import PESummaryNode
from .nodes.plot_node import PlotNode
from .nodes.sampling_node import SamplingNode

logger.name = "dingo_pipe"


def get_trigger_time_list(inputs):
    """Returns a list of GPS trigger times for each data segment"""
    # if (inputs.gaussian_noise or inputs.zero_noise) and inputs.trigger_time is None:
    #     trigger_times = [0] * inputs.n_simulation
    # elif (inputs.gaussian_noise or inputs.zero_noise) and isinstance(
    #     inputs.trigger_time, float
    # ):
    #     trigger_times = [inputs.trigger_time] * inputs.n_simulation
    if inputs.trigger_time is not None:
        trigger_times = [inputs.trigger_time]
    elif getattr(inputs, "gpstimes", None) is not None:
        start_times = inputs.gpstimes
        trigger_times = start_times + inputs.duration - inputs.post_trigger_duration
    else:
        raise BilbyPipeError("Unable to determine input trigger times from ini file")
    logger.info(f"Setting segment trigger-times {trigger_times}")
    return trigger_times


def get_parallel_list(inputs):
    if inputs.n_parallel == 1:
        return [""]
    else:
        return [f"part{idx}" for idx in range(inputs.n_parallel)]


def generate_dag(inputs, model_args):
    inputs = copy.deepcopy(inputs)
    dag = Dag(inputs)
    trigger_times = get_trigger_time_list(inputs)

    #
    # 1. Generate data for inference.
    #

    # Iterate over all generation nodes and store them in a list
    generation_node_list = []
    for idx, trigger_time in enumerate(trigger_times):
        kwargs = dict(trigger_time=trigger_time, idx=idx, dag=dag)
        if idx > 0:
            # Make all generation nodes depend on the 0th generation node
            # Ensures any cached files (e.g. the distance-marginalization
            # lookup table) are only built once.
            kwargs["parent"] = generation_node_list[0]
        generation_node = GenerationNode(inputs, **kwargs)
        generation_node_list.append(generation_node)

    #
    # 2. Sample the posterior using Dingo.
    #
    # Reconstruct the posterior density if necessary. This requires training a new
    # network and sampling from it.

    sampling_node_list = []
    for generation_node in generation_node_list:
        sampling_node = SamplingNode(
            inputs,
            generation_node=generation_node,
            dag=dag,
        )
        sampling_node_list.append(sampling_node)

    if inputs.importance_sample:
        #
        # 3. Generate new data for importance sampling **if different settings requested**.
        #
        # If injecting into simulated noise, be sure to use consistent noise realization.

        if len(inputs.importance_sampling_updates) > 0:
            # Iterate over all generation nodes and store them in a list
            importance_sampling_generation_node_list = []
            for idx, trigger_time in enumerate(trigger_times):
                kwargs = dict(trigger_time=trigger_time, idx=idx, dag=dag)
                if idx > 0:
                    # Make all generation nodes depend on the 0th generation node
                    # Ensures any cached files (e.g. the distance-marginalization
                    # lookup table) are only built once.
                    kwargs["parent"] = generation_node_list[0]
                generation_node = GenerationNode(inputs, importance_sampling=True, **kwargs)
                importance_sampling_generation_node_list.append(generation_node)
        else:
            importance_sampling_generation_node_list = generation_node_list

        #
        # 4. Importance sample
        #
        # If the phase is not present and phase marginalization is not being used, also
        # sample the phase synthetically. This adds between 1x and 50x to the cost of
        # importance sampling, depending on the waveform model. Indeed, IMRPhenomXPHM
        # waveform modes are much more expensive to generate than polarizations.

        merged_importance_sampling_node_list = []
        parallel_list = get_parallel_list(inputs)
        all_parallel_node_list = []
        for sampling_node, generation_node in zip(
            sampling_node_list, importance_sampling_generation_node_list
        ):
            parallel_node_list = []
            for parallel_idx in parallel_list:
                importance_sampling_node = ImportanceSamplingNode(
                    inputs,
                    sampling_node=sampling_node,
                    generation_node=generation_node,
                    parallel_idx=parallel_idx,
                    dag=dag,
                )
                parallel_node_list.append(importance_sampling_node)
                all_parallel_node_list.append(importance_sampling_node)

            if len(parallel_node_list) == 1:
                merged_importance_sampling_node_list.append(importance_sampling_node)
            else:
                # 4.(b) Recombine jobs into single Result.
                #       (Automatically calculates evidence.)
                merge_node = MergeNode(
                    inputs=inputs,
                    parallel_node_list=parallel_node_list,
                    dag=dag,
                )
                merged_importance_sampling_node_list.append(merge_node)

    else:
        merged_importance_sampling_node_list = sampling_node_list

    #
    # 5. Plotting
    #

    plot_nodes_list = []
    for merged_node in merged_importance_sampling_node_list:
        if inputs.plot_node_needed:
            plot_nodes_list.append(PlotNode(inputs, merged_node, dag=dag))

    #
    # 6. PESummary
    #

    if inputs.create_summary:
        # Add the waveform approximant to inputs, so that it can be fed to PESummary.
        inputs.waveform_approximant = model_args["waveform_approximant"]
        PESummaryNode(inputs, merged_importance_sampling_node_list, generation_node_list, dag=dag)

    dag.build()
    # create_overview(
    #     inputs,
    #     generation_node_list,
    #     all_parallel_node_list,
    #     merged_node_list,
    #     plot_nodes_list,
    # )



import copy
from typing import Iterable

from dingo.gw.domains import build_domain
from dingo.gw.gwutils import *
from dingo.gw.dataset import DingoDataset

HIGH_ASD_VALUE = 1.0


class ASDDataset(DingoDataset):
    """
    Dataset of amplitude spectral densities (ASDs). The ASDs are typically
    used for whitening strain data, and additionally passed as context to the
    neural density estimator.
    """

    dataset_type = "asd_dataset"

    def __init__(
        self,
        file_name=None,
        dictionary=None,
        ifos=None,
        precision=None,
        domain_update=None,
    ):
        """
        Parameters
        ----------
        file_name : str
            HDF5 file containing a dataset
        dictionary : dict
            Contains settings and data entries. The dictionary keys should be
            'settings', 'asds', and 'gps_times'.
        ifos : List[str]
            List of detectors used for dataset, e.g. ['H1', 'L1'].
            If not set, all available ones in the dataset are used.
        precision : str ('single', 'double')
            If provided, changes precision of loaded dataset.
        domain_update : dict
            If provided, update domain from existing domain using new settings.
        """
        self.precision = precision
        super().__init__(
            file_name=file_name,
            dictionary=dictionary,
            data_keys=["asds", "gps_times", "asd_parameterizations"],
        )

        if ifos is not None:
            for ifo in list(self.asds.keys()):
                if ifo not in ifos:
                    self.asds.pop(ifo)
                    self.gps_times.pop(ifo)

        self.domain = build_domain(self.settings["domain_dict"])
        if domain_update is not None:
            self.update_domain(domain_update)

        # Update dtypes if necessary
        if self.precision is not None:
            if self.precision == "single":
                for ifo, asd in self.asds.items():
                    self.asds[ifo] = asd.astype(np.float32, copy=False)
            elif self.precision == "double":
                for ifo, asd in self.asds.items():
                    self.asds[ifo] = asd.astype(np.float64, copy=False)
            else:
                raise TypeError(
                    'precision can only be changed to "single" or "double".'
                )

    @property
    def length_info(self):
        """The number of asd samples per detector."""
        return {key: len(val) for key, val in self.asds.items()}

    @property
    def gps_info(self):
        """Min/Max GPS time for each detector."""
        gps_info_dict = {}
        for key, val in self.gps_times.items():
            if not isinstance(val, Iterable):
                gps_info_dict[key] = val
            else:
                gps_info_dict[key] = (min(val), max(val))
        return gps_info_dict

    def update_domain(self, domain_update):
        """
        Update the domain based on new configuration. Also adjust data arrays to match
        the new domain.

        The ASD dataset provides ASDs in a particular domain. In Frequency domain,
        this is [0, domain._f_max]. In practice one may want to train a network based on
        slightly different domain settings, which corresponds to truncating the likelihood
        integral.

        This method provides functionality for that. It truncates the data below a
        new f_max, and sets the ASD below f_min to a large but finite value.

        Parameters
        ----------
        domain_update : dict
            Settings dictionary. Must contain a subset of the keys contained in
            domain_dict.
        """
        len_domain_original = len(self.domain)
        self.domain.update(domain_update)
        self.settings["domain"] = copy.deepcopy(self.domain.domain_dict)

        # truncate the dataset
        for ifo, asds in self.asds.items():

            # Is there a reason this check is needed? I would think that a dataset
            # should never be saved with this not matching.
            assert asds.shape[-1] == len_domain_original, (
                f"ASDs with shape {asds.shape[-1]} are not compatible"
                f"with the domain of length {len_domain_original}."
            )
            self.asds[ifo] = self.domain.update_data(
                asds,
                low_value=HIGH_ASD_VALUE,
            )

    def sample_random_asds(self):
        """
        Sample a random asd for each detector.
        Returns
        -------
        Dict with a random asd from the dataset for each detector.
        """
        return {k: v[np.random.choice(len(v), 1)[0]] for k, v in self.asds.items()}
